import asyncio
import decimal
import math
import time
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from datetime import datetime, date, timezone
from typing import Optional, Union, List, Tuple

import sentry_sdk
from dateutil.relativedelta import relativedelta
from sqlalchemy import MetaData, Table, Column, String, Integer, Numeric, Date, DateTime, Boolean, ARRAY, select, func, \
    distinct, case, text, literal_column, inspect, JSON
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.engine import Engine, Connection
from sqlalchemy.sql import FromClause, Selectable
from sqlalchemy.sql.elements import ColumnClause
from sqlalchemy.sql.expression import CTE, false, true, table as table_clause, column as column_clause
from sqlalchemy.types import Float

from .event import ProfilerEventHandler, DefaultProfilerEventHandler
from ..configuration import Configuration
from ..datasource import DataSource

HISTOGRAM_NUM_BUCKET = 50


class ProfileSubject:
    def __init__(self, table: str, schema: str = None, database: str = None, name: str = None, ref_id: str = None):
        self.table = table
        self.schema = schema
        self.database = database
        self.name = name if name else table
        self.ref_id = ref_id


def dtof(value: Union[int, float, decimal.Decimal]) -> Union[int, float]:
    """
    dtof is helpler function to transform decimal value to float. Decimal is not json serializable type.

    :param value:
    :return:
    """
    if isinstance(value, decimal.Decimal):
        return float(value)
    return value


def format_float(val: Union[int, float]) -> str:
    """
    from the float to human-readable format.

    :param val:
    :return:
    """

    if val == 0:
        return "0"

    base = math.floor(math.log10(abs(val)))
    if base < -2:
        return f"{val:.2e}"
    elif base < 0:
        return f"{val:.3f}"
    elif base < 3:
        return f"{val:.2f}"
    elif base < 6:
        return f"{val / (10 ** 3):.1f}K"
    elif base < 9:
        return f"{val / (10 ** 6):.1f}M"
    elif base < 12:
        return f"{val / (10 ** 9):.1f}B"
    elif base < 15:
        return f"{val / (10 ** 12):.1f}T"
    else:
        return f"{val / (10 ** 12):.0f}T"


def percentage(number, total):
    if number is None:
        return None
    if not total:
        return None
    return number / total


async def _run_in_executor(executor, func, *args):
    if executor:
        return await asyncio.get_running_loop().run_in_executor(executor, func, *args)
    else:
        return func(*args)


@dataclass
class CollectedMetadata:
    map_name_tables: dict
    profiled_tables: dict
    result: dict
    subjects: List[ProfileSubject]


class Profiler:
    """
    Profiler profile tables and columns by a sqlalchemy engine.
    """

    def __init__(
        self,
        data_source: DataSource,
        event_handler: ProfilerEventHandler = DefaultProfilerEventHandler(),
        config: Configuration = None
    ):
        self.data_source = data_source
        self.event_handler = event_handler
        self.config = config
        self.collected_metadata: Optional[CollectedMetadata] = None
        if self.data_source.threads > 1:
            self.executor = ThreadPoolExecutor(max_workers=self.data_source.threads)
        else:
            self.executor = None

    async def _fetch_metadata(self, subjects):
        futures = []
        map_name_tables = dict()
        total = len(subjects)
        completed = 0

        self.event_handler.handle_metadata_start()
        self.event_handler.handle_metadata_progress(total, completed)
        for subject in subjects:
            def _fetch_table_task(subject):
                engine = self.data_source.get_engine_by_database(subject.database)
                schema = subject.schema.lower() if subject.schema is not None else None
                table = None
                try:
                    table = Table(subject.table, MetaData(), autoload_with=engine, schema=schema)
                except BaseException as e:
                    # ignore the table metadata fetch error
                    sentry_sdk.capture_exception(e)
                    pass
                return subject.name, table

            future = _run_in_executor(self.executor, _fetch_table_task, subject)
            futures.append(future)

        for future in asyncio.as_completed(futures):
            name, table = await future
            map_name_tables[name] = table
            completed += 1
            self.event_handler.handle_metadata_progress(total, completed)
        self.event_handler.handle_metadata_end()

        return map_name_tables

    async def _profile(self, subjects: List[ProfileSubject] = None, *,
                       metadata_subjects: List[ProfileSubject] = None) -> dict:
        """
        profile all tables or specific table. With different column types, it would profile different metrics.

        The column can be categorized as these types
        - integer
        - numeric
        - string
        - datetime
        - boolean
        - Other

        :param subjects: optional, the tables to profile
        :return: the profile results
        """

        if self.collected_metadata is None:
            await self._collect_metadata(subjects, metadata_subjects)

        map_name_tables = self.collected_metadata.map_name_tables
        profiled_tables = self.collected_metadata.profiled_tables
        result = self.collected_metadata.result
        subjects = self.collected_metadata.subjects

        table_count = len(subjects)
        table_index = 0

        # Profiling
        if len(subjects) > 0:
            self.event_handler.handle_run_start(result)
            self.event_handler.handle_run_progress(result, table_count, table_index)

            for subject in subjects:
                name = subject.name
                table = map_name_tables.get(name)
                if table is None:
                    continue
                engine = self.data_source.get_engine_by_database(subject.database)
                table_profiler = TableProfiler(engine, self.executor, subject, table, self.event_handler, self.config)
                tresult = await table_profiler.profile()
                profiled_tables[name] = tresult
                table_index = table_index + 1
                self.event_handler.handle_run_progress(result, table_count, table_index)
            self.event_handler.handle_run_end(result)
        else:
            print("No models, seeds, sources to profile")

        return result

    async def _collect_metadata(self, subjects: List[ProfileSubject], metadata_subjects: List[ProfileSubject]):
        profiled_tables = {}
        result = {
            "tables": profiled_tables,
        }
        if subjects is None:
            subjects = []
            table_names = inspect(self.data_source.get_engine_by_database()).get_table_names()
            for table_name in table_names:
                subject = ProfileSubject(table_name)
                subjects.append(subject)

        # Fetch schema data
        map_name_tables = await self._fetch_metadata(metadata_subjects if metadata_subjects else subjects)
        map_name_tables = {k: v for k, v in map_name_tables.items() if v is not None}

        if metadata_subjects is None:
            # for compatible with non-dbt cases, we use subjects as the metadata_subjects
            metadata_subjects = subjects

        for subject in metadata_subjects:
            engine = self.data_source.get_engine_by_database(subject.database)
            table = map_name_tables.get(subject.name)
            if table is None:
                continue
            table_profiler = TableProfiler(engine, self.executor, subject, table, self.event_handler, self.config)
            tresult = await table_profiler.fetch_schema()
            profiled_tables[subject.name] = tresult

        self.collected_metadata = CollectedMetadata(map_name_tables=map_name_tables,
                                                    profiled_tables=profiled_tables,
                                                    result=result,
                                                    subjects=subjects)
        return self.collected_metadata

    def collect_metadata(self, metadata_subjects: List[ProfileSubject], subjects: List[ProfileSubject]):
        return asyncio.run(self._collect_metadata(subjects, metadata_subjects))

    def profile(self, subjects: List[ProfileSubject] = None, *, metadata_subjects: List[ProfileSubject] = None) -> dict:
        def job():
            return asyncio.run(self._profile(subjects, metadata_subjects=metadata_subjects))

        if not self.executor:
            return job()
        with self.executor:
            return job()


class TableProfiler:

    def __init__(
        self,
        engine: Engine,
        executor: ThreadPoolExecutor,
        subject: ProfileSubject,
        table: Table,
        event_handler: ProfilerEventHandler,
        config: Configuration
    ):
        self.engine = engine
        self.executor = executor
        self.subject = subject
        self.table = table
        self.event_handler = event_handler
        self.config = config

    def _get_candidate_columns(self) -> Tuple[Selectable, ColumnClause]:
        table = self.table
        if self.engine.url.get_backend_name() == 'bigquery':
            yield from self._get_candidate_columns_bigquery()
        else:
            for column in table.columns:
                yield table, column

    def _get_candidate_columns_bigquery(self) -> Tuple[Selectable, ColumnClause]:
        from sqlalchemy_bigquery import STRUCT, ARRAY

        # works around to fix the warning
        # SAWarning: UserDefinedType STRUCT(array=ARRAY(STRUCT(item1=String(), item2=String()))) will not produce a cache key because the ``cache_ok``
        # attribute is not set to True.  This can have significant performance implications including some performance degradations in comparison to prior SQLAlchemy versions.  Set this attribute to True if this type object's state is safe to use
        # in a cache key, or False to disable this warning. (Background on this error at: https://sqlalche.me/e/14/cprf)
        STRUCT.cache_ok = false
        ARRAY.cache_ok = false

        table = self.table
        cte_map = dict()

        cte_map[None] = select(
            text('*')
        ).select_from(
            table
        ).cte('t_')

        for column in table.columns:
            comps = column.name.split('.')
            name = comps[-1]
            prefix = '__'.join(comps[:-1]) if len(comps) > 1 else None
            selectable = cte_map.get(prefix)

            if isinstance(column.type, ARRAY):
                if isinstance(column.type.item_type, JSON) or isinstance(column.type.item_type, STRUCT):
                    # add array cte to cte map
                    cte_name = '__'.join(comps)
                    stmt = select(
                        literal_column(f"`{name}`.*", column.type)
                    ).select_from(
                        selectable
                    ).select_from(
                        text(f"unnest(`{selectable.name}`.`{name}`) as `{name}`")
                    ).cte("t_" + cte_name)
                    cte_map[cte_name] = stmt
                else:
                    # array cte
                    cte_name = '__'.join(comps)
                    selectable = select(
                        literal_column(f"`{name}`", column.type.item_type)
                    ).select_from(
                        selectable
                    ).select_from(
                        text(f"unnest(`{selectable.name}`.`{name}`) as `{name}`")
                    ).cte("t_" + cte_name)
                    yield selectable, literal_column(f"`{selectable.name}`.`{name}`", column.type).label(
                        column.name)
            elif isinstance(column.type, JSON) or isinstance(column.type, STRUCT):

                # add cte to ctep map
                cte_name = '__'.join(comps)
                stmt = select(
                    literal_column(f"`{selectable.name}`.`{name}`.*", column.type)
                ).select_from(
                    selectable
                ).cte("t_" + cte_name)
                cte_map[cte_name] = stmt
            else:
                yield selectable, literal_column(f"`{selectable.name}`.`{name}`", column.type).label(column.name)

    def _profile_table_metadata(self, result: dict):
        table = self.table
        row_count = created = last_altered = size_bytes = None
        with self.engine.connect() as conn:
            try:
                if self.engine.url.get_backend_name() == 'snowflake':
                    inspector = inspect(self.engine) if self.engine else None
                    default_schema = inspector.default_schema_name
                    metadata_table = table_clause('TABLES', column_clause("row_count"), column_clause("created"),
                                                  column_clause("last_altered"), column_clause("bytes"),
                                                  column_clause('table_schema'), column_clause('table_name'),
                                                  schema='INFORMATION_SCHEMA')
                    metadata_columns = {column.name: column for column in metadata_table.columns}
                    stmt = select(
                        metadata_columns['row_count'],
                        func.convert_timezone('UTC', metadata_columns['created']),
                        func.convert_timezone('UTC', metadata_columns['last_altered']),
                        metadata_columns['bytes']
                    ).select_from(metadata_table).where(metadata_columns['table_schema'] == str.upper(default_schema),
                                                        metadata_columns['table_name'] == str.upper(table.name))
                    row_count, created, last_altered, size_bytes = conn.execute(stmt).fetchone()
                    # datetime object transformation
                    created = created.isoformat()
                    last_altered = last_altered.isoformat()
                elif self.engine.url.get_backend_name() == 'bigquery':
                    dataset = self.engine.url.database
                    metadata_table = table_clause(f'{dataset}.__TABLES__', column_clause("row_count"),
                                                  column_clause("creation_time"), column_clause("last_modified_time"),
                                                  column_clause("size_bytes"), column_clause('table_id'))
                    metadata_columns = {column.name: column for column in metadata_table.columns}
                    stmt = select(
                        metadata_columns['row_count'],
                        metadata_columns['creation_time'],
                        metadata_columns['last_modified_time'],
                        metadata_columns['size_bytes']
                    ).select_from(metadata_table).where(metadata_columns['table_id'] == table.name)
                    row_count, created, last_altered, size_bytes = conn.execute(stmt).fetchone()
                    # timestamp transformation
                    created = datetime.fromtimestamp(created / 1000.0, timezone.utc).isoformat()
                    last_altered = datetime.fromtimestamp(last_altered / 1000.0, timezone.utc).isoformat()
                elif self.engine.url.get_backend_name() == 'redshift':
                    metadata_table = table_clause('SVV_TABLE_INFO', column_clause("tbl_rows"),
                                                  column_clause("size"), column_clause("table"))
                    metadata_columns = {column.name: column for column in metadata_table.columns}
                    stmt = select(
                        metadata_columns['tbl_rows'],
                        metadata_columns['size'],
                    ).select_from(metadata_table).where(metadata_columns['table'] == table.name)
                    row_count, size_mbytes = conn.execute(stmt).fetchone()
                    row_count = int(row_count)
                    size_bytes = size_mbytes * 1024
            except Exception:
                # table's metadata is optional except row_count
                pass
            finally:
                if row_count is None:
                    stmt = select(
                        func.count(),
                    ).select_from(table)
                    row_count, = conn.execute(stmt).fetchone()

        result['row_count'] = result['samples'] = row_count
        result['samples_p'] = 1

        if self.config:
            limit = self.config.profiler_config.get('table', {}).get('limit', 0)
            if row_count > limit > 0:
                result['samples'] = limit
                result['samples_p'] = percentage(limit, row_count)

        if created:
            result['created'] = created
        if last_altered:
            result['last_altered'] = last_altered
            freshness = datetime.now(timezone.utc) - datetime.fromisoformat(last_altered)
            result['freshness'] = int(freshness.total_seconds())
        if size_bytes:
            result['bytes'] = size_bytes

    def _profile_table_duplicate_rows(self, result: dict):
        table = self.table
        if not self.config:
            return
        if not self.config.profiler_config.get('table', {}).get('duplicateRows'):
            return

        limit = self.config.profiler_config.get('table', {}).get('limit', 0)
        columns = [column.label(f'_{column.name}') for column in table.columns]

        with self.engine.connect() as conn:
            if self.engine.url.get_backend_name() == 'snowflake':
                if limit <= 0:
                    cte = select(func.hash(*columns).label('h')).select_from(table).cte()
                else:
                    cte = select(func.hash(*columns).label('h')).select_from(table).limit(limit).cte()

                cte = select(
                    cte.c.h,
                    func.count().label('c')
                ).select_from(cte).group_by(cte.c.h).having(func.count() > 1).cte()
                stmt = select(func.sum(cte.c.c)).select_from(cte)
                duplicate_rows, = conn.execute(stmt).fetchone()
            else:
                if limit <= 0:
                    cte = select(
                        *columns,
                        func.count().label('c')
                    ).select_from(table).group_by(*columns).having(func.count() > 1).cte()
                else:
                    cte = select(*columns).select_from(table).limit(limit).cte()
                    columns = [column for column in cte.columns]
                    cte = select(
                        *columns,
                        func.count().label('c')
                    ).select_from(cte).group_by(*columns).having(func.count() > 1).cte()

                stmt = select(func.sum(cte.c.c)).select_from(cte)
                duplicate_rows, = conn.execute(stmt).fetchone()

            samples = result['samples']
            duplicate_rows = duplicate_rows if duplicate_rows is not None else 0

            result['duplicate_rows'] = duplicate_rows
            result['duplicate_rows_p'] = percentage(duplicate_rows, samples)

    async def _profile_table(self, result):
        await _run_in_executor(self.executor, self._profile_table_metadata, result)
        await _run_in_executor(self.executor, self._profile_table_duplicate_rows, result)

    async def _profile_column(self, result, table_name, table: Table, column: Column) -> dict:
        column_name = column.name
        column_result, profiler = await self._create_column_metadata_and_profiler(table, column)

        self.event_handler.handle_column_start(table_name, column_name)

        profile_start = time.perf_counter()
        profile_result = await _run_in_executor(self.executor, profiler.profile)
        profile_end = time.perf_counter()
        duration = profile_end - profile_start

        column_result.update(profile_result)
        column_result["profile_duration"] = f"{duration:.2f}"
        column_result["elapsed_milli"] = int(duration * 1000)

        self.event_handler.handle_column_end(table_name, column_name, column_result)
        result['columns'][column_name] = column_result

    async def _create_column_metadata_and_profiler(self, table, column):
        profiler_config = self.config.profiler_config if self.config else {}
        column_type = column.type
        schema_type = str(column.type)
        if isinstance(column_type, ARRAY) and column_type.item_type is not None:
            column_type = column_type.item_type
            schema_type = f"ARRAY<{column_type}>"
        if isinstance(column_type, String):
            # VARCHAR
            # CHAR
            # TEXT
            # CLOB
            generic_type = "string"
            profiler = StringColumnProfiler(self.engine, profiler_config, table, column)
        elif isinstance(column_type, Integer):
            # INTEGER
            # BIGINT
            # SMALLINT
            generic_type = "integer"
            profiler = NumericColumnProfiler(self.engine, profiler_config, table, column, is_integer=True)
        elif isinstance(column_type, Numeric):
            # NUMERIC
            # DECIMAL
            # FLOAT
            generic_type = "numeric"
            profiler = NumericColumnProfiler(self.engine, profiler_config, table, column, is_integer=False)
        elif isinstance(column_type, Date) or isinstance(column_type, DateTime) or \
            (self.engine.url.get_backend_name() == 'snowflake' and str(column_type).startswith('TIMESTAMP')):
            # DATE
            # DATETIME
            # TIMEZONE_NTZ
            generic_type = "datetime"
            profiler = DatetimeColumnProfiler(self.engine, profiler_config, table, column)
        elif isinstance(column_type, Boolean):
            # BOOLEAN
            generic_type = "boolean"
            profiler = BooleanColumnProfiler(self.engine, profiler_config, table, column)
        elif isinstance(column_type, UUID):
            generic_type = "other"
            profiler = UUIDColumnProfiler(self.engine, profiler_config, table, column)
        else:
            generic_type = "other"
            profiler = BaseColumnProfiler(self.engine, profiler_config, table, column)
        column_result = {
            "name": column.name,
            "type": generic_type,
            "schema_type": schema_type,
        }
        return column_result, profiler

    async def profile(self) -> dict:
        subject = self.subject
        name = subject.name
        self.event_handler.handle_table_start(name)
        candidate_columns = list(self._get_candidate_columns())
        col_index = 0
        col_count = len(candidate_columns)
        columns = {}
        result = {
            "name": name,
            "row_count": 0,
            "samples": 0,
            "samples_p": None,
            "col_count": col_count,
            "duplicate_rows": None,
            "duplicate_rows_p": None,
            "columns": columns,
        }
        if subject.ref_id:
            result['ref_id'] = subject.ref_id
        futures = []
        profile_start = time.perf_counter()

        self.event_handler.handle_table_progress(name, result, col_count, col_index)

        # Profile table
        future = asyncio.create_task(self._profile_table(result))
        futures.append(future)

        # Profile columns
        for selectable, column in candidate_columns:
            columns[column.name] = None
            future = asyncio.create_task(self._profile_column(result, name, selectable, column))
            futures.append(future)

        total = len(futures)
        completed = 0
        self.event_handler.handle_table_progress(name, result, total, completed)

        for future in asyncio.as_completed(futures):
            await future
            completed += 1
            self.event_handler.handle_table_progress(name, result, total, completed)

        for column_result in columns.values():
            column_result['total'] = result['row_count']
            column_result['samples_p'] = result['samples_p']

        profile_end = time.perf_counter()
        duration = profile_end - profile_start
        result["profile_duration"] = f"{duration:.2f}"
        result["elapsed_milli"] = int(duration * 1000)

        self.event_handler.handle_table_end(name, result)
        return result

    async def fetch_schema(self) -> dict:
        subject = self.subject
        name = subject.name
        candidate_columns = list(self._get_candidate_columns())
        col_count = len(candidate_columns)
        columns = {}
        result = {
            "name": name,
            "col_count": col_count,
            "columns": columns,
        }

        if subject.ref_id:
            result['ref_id'] = subject.ref_id

        futures = []

        # Profile columns
        for selectable, column in candidate_columns:
            columns[column.name] = None
            future = asyncio.create_task(self._create_column_metadata_and_profiler(selectable, column))
            futures.append(future)

        for future in asyncio.as_completed(futures):
            column_result, _ = await future
            columns[column_result['name']] = column_result
        return result


class BaseColumnProfiler:
    """
    The base class of the column profiler. It will automatically profile the metrics according to the schema type
    """

    def __init__(self, engine: Engine, config: dict, table: Table, column: Column):
        self.engine = engine
        self.config = config
        self.table = table
        self.column = column

    def _get_database_backend(self) -> str:
        """
        Helper function to return the sqlalchemy engine backend
        :return:
        """
        return self.engine.url.get_backend_name()

    def _get_limited_table_cte(self):
        t = self.table
        c = self.column
        if not self.config:
            return t, c

        limit = self.config.get('table', {}).get('limit', 0)
        if limit <= 0:
            return t, c
        else:
            cte = select(c.label('c')).select_from(t).limit(limit).cte()
            return cte, cte.c.c

    def _get_table_cte(self) -> CTE:
        """
        Get the CTE to normalize the
        - table name
        - column name as column "c"
        - (Optional) Remove the invalid data.

        Columns
        - "c": the transformed valid to use data
            null: if the column value is null or invalid
            otherwise: orginal column value or transformed value.
        - "orig": the original column

        so that

        valid       = count(c)
        non_nulls   = count(orig)
        invalids    = non_nulls - valid

        :return: CTE
        """
        t, c = self._get_limited_table_cte()

        return select(c.label("c")).select_from(t).cte()

    def profile(self) -> dict:
        """
        Profile a column

        :return: the profiling result. The result dict is json serializable
        """

        with self.engine.connect() as conn:
            cte = self._get_table_cte()
            stmt = select(
                func.count().label("_total"),
                func.count(cte.c.c).label("_non_nulls"),
            )
            result = conn.execute(stmt).fetchone()
            _total, _non_nulls, = result
            _nulls = _total - _non_nulls
            _valid = _non_nulls

            return {
                'total': None,
                'samples': _total,
                'samples_p': None,
                'non_nulls': _non_nulls,
                'non_nulls_p': percentage(_non_nulls, _total),
                'nulls': _nulls,
                'nulls_p': percentage(_nulls, _total),
                'valids': _valid,
                'valids_p': percentage(_valid, _total),
                'invalids': 0,
                'invalids_p': 0
            }


class StringColumnProfiler(BaseColumnProfiler):
    def __init__(self, engine: Engine, config: dict, table: Table, column: Column):
        super().__init__(engine, config, table, column)

    def _get_table_cte(self) -> CTE:
        t, c = self._get_limited_table_cte()
        if self._get_database_backend() != 'sqlite':
            cte = select(
                c.label("c"),
                c.label("orig")
            ).select_from(t).cte()
        else:
            cte = select(
                case(
                    (func.typeof(c) == 'blob', None),
                    else_=c
                ).label("c"),
                c.label("orig"),
            ).select_from(t).cte()
        cte = select(
            cte.c.c,
            func.length(cte.c.c).label("len"),
            cte.c.orig
        ).select_from(cte).cte()
        cte = select(
            cte.c.c,
            cte.c.len,
            case((cte.c.len == 0, 1), else_=None).label("zero_length"),
            cte.c.orig
        ).select_from(cte).cte()
        return cte

    def profile(self):
        with self.engine.connect() as conn:
            cte = self._get_table_cte()

            columns = [
                func.count().label("_total"),
                func.count(cte.c.orig).label("_non_nulls"),
                func.count(cte.c.c).label("_valids"),
                func.count(cte.c.zero_length).label("_zero_length"),
                func.count(distinct(cte.c.c)).label("_distinct"),
                func.avg(cte.c.len).label("_avg"),
                func.min(cte.c.len).label("_min"),
                func.max(cte.c.len).label("_max"),
            ]

            if self._get_database_backend() == 'sqlite':
                columns.append((func.count(cte.c.len) * func.sum(
                    func.cast(cte.c.len, Float) * func.cast(cte.c.len, Float)) - func.sum(cte.c.len) * func.sum(
                    cte.c.len)) / ((func.count(cte.c.len) - 1) * func.count(cte.c.len)).label('_variance'))
                stmt = select(*columns)
                result = conn.execute(stmt).fetchone()
                _total, _non_nulls, _valids, _zero_length, _distinct, _avg, _min, _max, _variance = result
                _stddev = None
                if _variance is not None:
                    _stddev = math.sqrt(_variance)
            else:
                columns.append(func.stddev(cte.c.len).label("_stddev"))
                stmt = select(*columns)
                result = conn.execute(stmt).fetchone()
                _total, _non_nulls, _valids, _zero_length, _distinct, _avg, _min, _max, _stddev = result

            _nulls = _total - _non_nulls
            _invalids = _non_nulls - _valids
            _non_zero_length = _valids - _zero_length
            _min = dtof(_min)
            _max = dtof(_max)
            _avg = dtof(_avg)
            _stddev = dtof(_stddev)

            result = {
                'total': None,
                'samples': _total,
                'samples_p': None,
                'non_nulls': _non_nulls,
                'non_nulls_p': percentage(_non_nulls, _total),
                'nulls': _nulls,
                'nulls_p': percentage(_nulls, _total),
                'valids': _valids,
                'valids_p': percentage(_valids, _total),
                'invalids': _invalids,
                'invalids_p': percentage(_invalids, _total),
                'zero_length': _zero_length,
                'zero_length_p': percentage(_zero_length, _total),
                'non_zero_length': _non_zero_length,
                'non_zero_length_p': percentage(_non_zero_length, _total),

                'distinct': _distinct,
                'distinct_p': percentage(_distinct, _valids),
                'min': _min,
                'min_length': _min,
                'max': _max,
                'max_length': _max,
                'avg': _avg,
                'avg_length': _avg,
                'stddev': _stddev,
                'stddev_length': _stddev,
            }

            # uniqueness
            _non_duplicates = profile_non_duplicate(conn, cte, cte.c.c)
            _duplicates = _valids - _non_duplicates
            result.update({
                "duplicates": _duplicates,
                "duplicates_p": percentage(_duplicates, _valids),
                "non_duplicates": _non_duplicates,
                "non_duplicates_p": percentage(_non_duplicates, _valids),
            })

            # top k
            topk = None
            if _valids > 0:
                topk = profile_topk(conn, cte.c.c)
            result['topk'] = topk

            # histogram of string length
            histogram = None
            if _valids > 0:
                histogram = profile_histogram(conn, cte, cte.c.len, _min, _max, True)
            result['histogram'] = histogram
            result['histogram_length'] = histogram

            return result


class NumericColumnProfiler(BaseColumnProfiler):
    is_integer: bool

    def __init__(self, engine: Engine, config: dict, table: Table, column: Column, is_integer: bool):
        super().__init__(engine, config, table, column)
        self.is_integer = is_integer

    def _get_table_cte(self) -> CTE:
        t, c = self._get_limited_table_cte()
        if self._get_database_backend() != 'sqlite':
            cte = select(
                c.label("c"),
                c.label("orig")
            ).select_from(t).cte()
        else:
            cte = select(
                case(
                    (func.typeof(c) == 'text', None),
                    (func.typeof(c) == 'blob', None),
                    else_=c
                ).label("c"),
                c.label("orig")
            ).select_from(t).cte(name="T")
        cte = select(
            cte.c.c,
            case((cte.c.c == 0, 1), else_=None).label("zero"),
            case((cte.c.c < 0, 1), else_=None).label("negative"),
            cte.c.orig
        ).select_from(cte).cte()
        return cte

    def profile(self):
        with self.engine.connect() as conn:
            cte = self._get_table_cte()

            columns = [
                func.count().label("_total"),
                func.count(cte.c.orig).label("_non_nulls"),
                func.count(cte.c.c).label("_valids"),
                func.count(cte.c.zero).label("_zeros"),
                func.count(cte.c.negative).label("_negatives"),
                func.count(distinct(cte.c.c)).label("_distinct"),
                func.sum(func.cast(cte.c.c, Float)).label("_sum"),
                func.avg(cte.c.c).label("_avg"),
                func.min(cte.c.c).label("_min"),
                func.max(cte.c.c).label("_max"),
            ]

            if self._get_database_backend() == 'sqlite':
                columns.append((func.count(cte.c.c) * func.sum(
                    func.cast(cte.c.c, Float) * func.cast(cte.c.c, Float)) - func.sum(cte.c.c) * func.sum(cte.c.c)) / (
                                   (func.count(cte.c.c) - 1) * func.count(cte.c.c)).label('_variance'))
                stmt = select(*columns)
                result = conn.execute(stmt).fetchone()
                _total, _non_nulls, _valids, _zeros, _negatives, _distinct, _sum, _avg, _min, _max, _variance = result
                _stddev = None
                if _variance is not None:
                    _stddev = math.sqrt(_variance)
            else:
                columns.append(func.stddev(func.cast(cte.c.c, Float)).label("_stddev"))
                stmt = select(*columns)
                result = conn.execute(stmt).fetchone()
                _total, _non_nulls, _valids, _zeros, _negatives, _distinct, _sum, _avg, _min, _max, _stddev = result

            _nulls = _total - _non_nulls
            _invalids = _non_nulls - _valids
            _positives = _valids - _zeros - _negatives
            _sum = dtof(_sum)
            _min = dtof(_min)
            _max = dtof(_max)
            _avg = dtof(_avg)
            _stddev = dtof(_stddev)

            result = {
                'total': None,
                'samples': _total,
                'samples_p': None,
                'non_nulls': _non_nulls,
                'non_nulls_p': percentage(_non_nulls, _total),
                'nulls': _nulls,
                'nulls_p': percentage(_nulls, _total),
                'valids': _valids,
                'valids_p': percentage(_valids, _total),
                'invalids': _invalids,
                'invalids_p': percentage(_invalids, _total),
                'zeros': _zeros,
                'zeros_p': percentage(_zeros, _total),
                'negatives': _negatives,
                'negatives_p': percentage(_negatives, _total),
                'positives': _positives,
                'positives_p': percentage(_positives, _total),

                'distinct': _distinct,
                'distinct_p': percentage(_distinct, _valids),
                'min': _min,
                'max': _max,
                'sum': _sum,
                'avg': _avg,
                'stddev': _stddev,
            }

            # uniqueness
            _non_duplicates = profile_non_duplicate(conn, cte, cte.c.c)
            _duplicates = _valids - _non_duplicates
            result.update({
                "duplicates": _duplicates,
                "duplicates_p": percentage(_duplicates, _valids),
                "non_duplicates": _non_duplicates,
                "non_duplicates_p": percentage(_non_duplicates, _valids),
            })

            # histogram
            histogram = None
            if _valids > 0 and math.isfinite(_min) and math.isfinite(_max):
                histogram = profile_histogram(conn, cte, cte.c.c, _min, _max, self.is_integer)
            result['histogram'] = histogram

            # quantile
            quantile = {}
            if _valids > 0 and math.isfinite(_min) and math.isfinite(_max):
                quantile = self._profile_quantile(conn, cte, cte.c.c, _valids)
            result.update({
                'p5': quantile.get('p5'),
                'p25': quantile.get('p25'),
                'p50': quantile.get('p50'),
                'p75': quantile.get('p75'),
                'p95': quantile.get('p95'),
            })

            # top k (integer only)
            if self.is_integer:
                topk = None
                if _valids > 0:
                    topk = profile_topk(conn, cte.c.c)
                result["topk"] = topk

            return result

    def _profile_quantile_via_window_function(
        self,
        conn: Connection,
        table: FromClause,
        column: ColumnClause,
        total: int
    ):
        # with t as (
        #   select
        #     column as c,
        #     ntile(20) over (order by column) as n
        #   from table
        # )
        # select n, min(c) from t group by n order by n
        n_bucket = total if total < 100 else 100
        t = select(
            column.label("c"),
            func.ntile(n_bucket).over(order_by=column).label("n")
        ).where(column.isnot(None)).select_from(table).cte()
        stmt = select(t.c.n, func.min(t.c.c)).group_by(t.c.n).order_by(t.c.n)
        result = conn.execute(stmt)
        quantile = []
        for row in result:
            n, v = row
            quantile.append(v)
        return {
            'p5': dtof(quantile[5 * n_bucket // 100]),
            'p25': dtof(quantile[25 * n_bucket // 100]),
            'p50': dtof(quantile[50 * n_bucket // 100]),
            'p75': dtof(quantile[75 * n_bucket // 100]),
            'p95': dtof(quantile[95 * n_bucket // 100]),
        }

    def _profile_quantile_via_query_one_by_one(
        self,
        conn: Connection,
        table: FromClause,
        column: ColumnClause,
        total: int
    ):
        # Query for each quantile
        def ntile(n):
            offset = n * total // 100

            stmt = select(
                column
            ).select_from(
                table
            ).where(
                column.isnot(None)
            ).order_by(
                column
            ).offset(
                offset
            ).limit(1)
            result, = conn.execute(stmt).fetchone()
            return dtof(result)

        return {
            'p5': ntile(5),
            'p25': ntile(25),
            'p50': ntile(50),
            'p75': ntile(75),
            'p95': ntile(95),
        }

    def _profile_quantile(
        self,
        conn: Connection,
        table: FromClause,
        column: ColumnClause,
        total: int
    ) -> dict:
        """

        :param conn:
        :param table: a
        :param column:
        :param total:
        :return:
        """
        backend = self._get_database_backend()

        if backend == 'sqlite':
            import sqlite3
            version = sqlite3.sqlite_version.split(".")

            major = int(version[0]) if len(version) >= 2 else 0
            minor = int(version[1]) if len(version) >= 2 else 0

            if major > 3 or (major == 3 and minor >= 25):
                # use window function if sqlite version >= 3.25.0
                # see https://www.sqlite.org/windowfunctions.html

                return self._profile_quantile_via_window_function(conn, table, column, total)
            else:
                return self._profile_quantile_via_query_one_by_one(conn, table, column, total)
        elif backend == 'duckdb':
            selects = [
                func.approx_quantile(column, literal_column(f"{percentile}")) for percentile in
                [0.05, 0.25, 0.5, 0.75, 0.95]
            ]
        elif backend == 'bigquery':
            # BigQuery does not support WITHIN, change to use over
            #   Ref: https://github.com/great-expectations/great_expectations/blob/develop/great_expectations/dataset/sqlalchemy_dataset.py#L1019:9
            selects = [
                func.percentile_disc(column, percentile).over() for percentile in [0.05, 0.25, 0.5, 0.75, 0.95]
            ]
        elif backend == 'redshift':
            # ref: https://docs.aws.amazon.com/redshift/latest/dg/r_APPROXIMATE_PERCENTILE_DISC.html
            selects = [
                func.approximate_percentile_disc(percentile).within_group(column) for percentile in
                [0.05, 0.25, 0.5, 0.75, 0.95]
            ]
        elif backend == 'awsathena':
            selects = [
                func.approx_percentile(column, percentile) for percentile in
                [0.05, 0.25, 0.5, 0.75, 0.95]
            ]
        else:
            # https://docs.sqlalchemy.org/en/14/core/functions.html#sqlalchemy.sql.functions.percentile_disc
            #
            # select
            #     percentile_disc(0.05) within group (order by column),
            #     percentile_disc(0.25) within group (order by column),
            #     percentile_disc(0.5) within group (order by column),
            #     percentile_disc(0.75) within group (order by column),
            #     percentile_disc(0.95) within group (order by column)
            # from table
            selects = [
                func.percentile_disc(percentile).within_group(column) for percentile in [0.05, 0.25, 0.5, 0.75, 0.95]
            ]

        stmt = select(*selects).select_from(table)
        result = conn.execute(stmt).fetchone()
        return {
            'p5': dtof(result[0]),
            'p25': dtof(result[1]),
            'p50': dtof(result[2]),
            'p75': dtof(result[3]),
            'p95': dtof(result[4]),
        }

    def _profile_histogram(
        self,
        conn: Connection,
        table: FromClause,
        column: ColumnClause,
        min: Union[int, float],
        max: Union[int, float],
        is_integer: bool,
        num_buckets: int = HISTOGRAM_NUM_BUCKET
    ) -> dict:
        if is_integer:
            # min=0, max=50, num_buckets=50  => interval=1, num_buckets=51
            # min=0, max=70, num_buckets=50  => interval=2, num_buckets=36
            # min=0, max=100, num_buckets=50 => interval=2, num_buckets=51
            interval = math.ceil((max - min) / num_buckets) if max > min else 1
            num_buckets = math.ceil((max - min + 1) / interval)
        else:
            interval = (max - min) / num_buckets if max > min else 1

        cases = []
        for i in range(num_buckets):
            bound = min + interval * (i + 1)
            if i != num_buckets - 1:
                cases += [(column < bound, i)]
            else:
                cases += [(column < bound + interval / 100, i)]

        cte_with_bucket = select(
            column.label("c"),
            case(*cases, else_=None).label("bucket")
        ).select_from(
            table
        ).where(
            column.isnot(None)
        ).cte()

        stmt = select(
            cte_with_bucket.c.bucket,
            func.count().label("_count")
        ).group_by(
            cte_with_bucket.c.bucket
        ).order_by(
            cte_with_bucket.c.bucket
        )

        result = conn.execute(stmt)

        counts = []
        labels = []
        bin_edges = []
        for i in range(num_buckets):
            if is_integer:
                start = min + i * interval
                end = min + (i + 1) * interval
                if interval == 1:
                    label = f"{start}"
                else:
                    label = f"{start} _ {end}"
            else:
                if interval >= 1:
                    start = min + i * interval
                    end = min + (i + 1) * interval
                else:
                    start = min + i / (1 / interval)
                    end = min + (i + 1) / (1 / interval)

                label = f"{format_float(start)} _ {format_float(end)}"

            labels.append(label)
            counts.append(0)
            bin_edges.append(start)
            if i == num_buckets - 1:
                bin_edges.append(end)

        for row in result:
            _bucket, v = row
            if _bucket is None:
                continue
            counts[int(_bucket)] = v
        return {
            "type": "histogram",
            "labels": labels,
            "counts": counts,
            "bin_edges": bin_edges,
        }


class DatetimeColumnProfiler(BaseColumnProfiler):
    def __init__(self, engine: Engine, config: dict, table: Table, column: Column):
        super().__init__(engine, config, table, column)

    def _get_table_cte(self) -> CTE:
        t, c = self._get_limited_table_cte()
        if self._get_database_backend() != 'sqlite':
            cte = select(
                c.label("c"),
                c.label("orig")
            ).select_from(t).cte()
        else:
            cte = select(
                case(
                    (func.typeof(c) == 'text', func.datetime(c)),
                    (func.typeof(c) == 'integer', func.datetime(c, 'unixepoch')),
                    (func.typeof(c) == 'real', func.datetime(c, 'unixepoch')),
                    else_=None
                ).label("c"),
                c.label("orig"),
            ).select_from(t).cte()
        return cte

    def profile(self):
        with self.engine.connect() as conn:
            cte = self._get_table_cte()

            stmt = select(
                func.count().label("_total"),
                func.count(cte.c.orig).label("_non_nulls"),
                func.count(cte.c.c).label("_valids"),
                func.count(distinct(cte.c.c)).label("_distinct"),
                func.min(cte.c.c).label("_min"),
                func.max(cte.c.c).label("_max"),
            )
            result = conn.execute(stmt).fetchone()
            _total, _non_nulls, _valids, _distinct, _min, _max = result
            _nulls = _total - _non_nulls
            _invalids = _non_nulls - _valids

            if self._get_database_backend() == 'sqlite':
                if isinstance(self.column.type, Date):
                    _min = datetime.fromisoformat(_min).date() if isinstance(_min, str) else _min
                    _max = datetime.fromisoformat(_max).date() if isinstance(_max, str) else _max
                else:
                    _min = datetime.fromisoformat(_min) if isinstance(_min, str) else _min
                    _max = datetime.fromisoformat(_max) if isinstance(_max, str) else _max

            result = {
                'total': None,
                'samples': _total,
                'samples_p': None,
                'non_nulls': _non_nulls,
                'non_nulls_p': percentage(_non_nulls, _total),
                'nulls': _nulls,
                'nulls_p': percentage(_nulls, _total),
                'valids': _valids,
                'valids_p': percentage(_valids, _total),
                'invalids': _invalids,
                'invalids_p': percentage(_invalids, _total),
                'distinct': _distinct,
                'distinct_p': percentage(_distinct, _valids),
                'min': _min.isoformat() if _min is not None else None,
                'max': _max.isoformat() if _max is not None else None,
            }

            # uniqueness
            _non_duplicates = profile_non_duplicate(conn, cte, cte.c.c)
            _duplicates = _valids - _non_duplicates
            result.update({
                "duplicates": _duplicates,
                "duplicates_p": percentage(_duplicates, _valids),
                "non_duplicates": _non_duplicates,
                "non_duplicates_p": percentage(_non_duplicates, _valids),
            })

            # histogram
            histogram = None
            _type = None
            if _min and _max:
                histogram, _type = self._profile_histogram(conn, cte, cte.c.c, _min, _max)
            result['histogram'] = histogram

            return result

    def _profile_histogram(
        self,
        conn: Connection,
        table: FromClause,
        column: ColumnClause,
        min: Union[date, datetime],
        max: Union[date, datetime]
    ) -> Tuple[dict, str]:
        """
        Profile the histogram of a datetime column. There are three way to create bins of the histogram


        :param conn:
        :param table:
        :param column:
        :param min:
        :param max:
        :return:
        """

        # if self._get_database_backend() == 'sqlite':
        #     min = datetime.fromisoformat(min).date() if min is not None else min
        #     max = datetime.fromisoformat(max).date() if max is not None else max
        # else:
        #     if isinstance(min, datetime):
        #         min = min.date()
        #     if isinstance(max, datetime):
        #         max = max.date()

        _type = None
        histogram = {
            "labels": [],
            "counts": [],
            "bin_edges": [],
        }

        def date_trunc(*args):
            if self._get_database_backend() == 'sqlite':
                if args[0] == "YEAR":
                    return func.strftime("%Y-01-01", args[1])
                elif args[0] == "MONTH":
                    return func.strftime("%Y-%m-01", args[1])
                else:
                    return func.strftime("%Y-%m-%d", args[1])
            elif self._get_database_backend() == 'bigquery':
                date_expression = args[1]
                date_part = args[0]
                return func.date_trunc(date_expression, text(date_part))
            else:
                return func.date_trunc(*args)

        days_delta = (max - min).days
        if days_delta > 365 * 4:
            _type = "yearly"
            dmin = date(min.year, 1, 1)
            if max.year < 3000:
                dmax = date(max.year, 1, 1) + relativedelta(years=+1)
            else:
                dmax = date(3000, 1, 1)
            interval_years = math.ceil((dmax.year - dmin.year) / 50)
            interval = relativedelta(years=+interval_years)
            num_buckets = math.ceil((dmax.year - dmin.year) / interval.years)

            cte = select(date_trunc("YEAR", column).label("d")).select_from(table).cte()
        elif days_delta > 60:
            _type = "monthly"
            interval = relativedelta(months=+1)
            dmin = date(min.year, min.month, 1)
            dmax = date(max.year, max.month, 1) + interval
            period = relativedelta(dmax, dmin)
            num_buckets = (period.years * 12 + period.months)
            cte = select(date_trunc("MONTH", column).label("d")).select_from(table).cte()
        else:
            _type = "daily"
            interval = relativedelta(days=+1)
            dmin = date(min.year, min.month, min.day)
            dmax = date(max.year, max.month, max.day) + interval
            num_buckets = (dmax - dmin).days
            cte = select(date_trunc("DAY", column).label("d")).select_from(table).cte()

        stmt = select(
            cte.c.d,
            func.count(cte.c.d).label("_count")
        ).group_by(
            cte.c.d
        ).order_by(
            cte.c.d
        )

        result = conn.execute(stmt)

        for i in range(num_buckets):
            label = f"{dmin + i * interval} - {dmin + (i + 1) * interval}"
            histogram["labels"].append(label)
            histogram["bin_edges"].append(str(dmin + i * interval))
            histogram["counts"].append(0)
        histogram["bin_edges"].append(str(dmin + num_buckets * interval))

        for row in result:
            date_truncated, v = row
            if date_truncated is None:
                continue
            elif isinstance(date_truncated, str):
                date_truncated = date.fromisoformat(date_truncated)
            elif isinstance(date_truncated, datetime):
                date_truncated = date_truncated.date()

            for i in range(num_buckets):
                date_edge = date.fromisoformat(histogram["bin_edges"][i + 1])
                if date_truncated < date_edge:
                    histogram["counts"][i] += v
                    break

        return histogram, _type


class BooleanColumnProfiler(BaseColumnProfiler):
    def __init__(self, engine: Engine, config: dict, table: Table, column: Column):
        super().__init__(engine, config, table, column)

    def _get_table_cte(self) -> CTE:
        t, c = self._get_limited_table_cte()
        if self._get_database_backend() != 'sqlite':
            cte = select(
                c.label("c"),
                c.label("orig")
            ).select_from(t).cte()
        else:
            cte = select(
                case(
                    (c == true(), c),
                    (c == false(), c),
                    else_=None
                ).label("c"),
                c.label("orig"),
            ).cte()
        cte = select(
            cte.c.c,
            case((cte.c.c == true(), 1), else_=None).label("true_count"),
            cte.c.orig
        ).select_from(cte).cte()
        return cte

    def profile(self):
        cte = self._get_table_cte()

        with self.engine.connect() as conn:
            stmt = select(
                func.count().label("_total"),
                func.count(cte.c.orig).label("_non_nulls"),
                func.count(cte.c.c).label("_valids"),
                func.count(cte.c.true_count).label("_trues"),
                func.count(distinct(cte.c.c)).label("_distinct"),
            ).select_from(cte)
            result = conn.execute(stmt).fetchone()
            _total, _non_nulls, _valids, _trues, _distinct = result
            _nulls = _total - _non_nulls
            _invalids = _non_nulls - _valids
            _falses = _valids - _trues

            result = {
                'total': None,
                'samples': _total,
                'samples_p': None,
                'non_nulls': _non_nulls,
                'non_nulls_p': percentage(_non_nulls, _total),
                'nulls': _nulls,
                'nulls_p': percentage(_nulls, _total),
                'valids': _valids,
                'valids_p': percentage(_valids, _total),
                'invalids': _invalids,
                'invalids_p': percentage(_invalids, _total),
                'trues': _trues,
                'trues_p': percentage(_trues, _total),
                'falses': _falses,
                'falses_p': percentage(_falses, _total),
                'distinct': _distinct,
                'distinct_p': percentage(_distinct, _valids),
            }

            return result


class UUIDColumnProfiler(BaseColumnProfiler):
    def __init__(self, engine: Engine, config: dict, table: Table, column: Column):
        super().__init__(engine, config, table, column)

    def _get_table_cte(self) -> CTE:
        t, c = self._get_limited_table_cte()
        return select(c.label("c")).select_from(t).cte()

    def profile(self):
        with self.engine.connect() as conn:
            cte = self._get_table_cte()

            columns = [
                func.count().label("_total"),
                func.count(cte.c.c).label("_non_nulls"),
                func.count(distinct(cte.c.c)).label("_distinct"),
            ]

            stmt = select(*columns)
            result = conn.execute(stmt).fetchone()
            _total, _non_nulls, _distinct = result

            _nulls = _total - _non_nulls
            _valids = _non_nulls
            _invalids = _non_nulls - _valids

            result = {
                'total': None,
                'samples': _total,
                'samples_p': None,
                'non_nulls': _non_nulls,
                'non_nulls_p': percentage(_non_nulls, _total),
                'nulls': _nulls,
                'nulls_p': percentage(_nulls, _total),
                'valids': _valids,
                'valids_p': percentage(_valids, _total),
                'invalids': _invalids,
                'invalids_p': percentage(_invalids, _total),
                'distinct': _distinct,
                'distinct_p': percentage(_distinct, _valids),
            }

            # uniqueness
            _non_duplicates = profile_non_duplicate(conn, cte, cte.c.c)
            _duplicates = _valids - _non_duplicates
            result.update({
                "duplicates": _duplicates,
                "duplicates_p": percentage(_duplicates, _valids),
                "non_duplicates": _non_duplicates,
                "non_duplicates_p": percentage(_non_duplicates, _valids),
            })

            # top k
            topk = None
            if _valids > 0:
                topk = profile_topk(conn, func.cast(cte.c.c, String))
            result['topk'] = topk

            return result


def profile_topk(conn, expr, k=50) -> dict:
    stmt = select(
        expr,
        func.count().label("_count")
    ).where(
        expr.isnot(None)
    ).group_by(
        expr
    ).order_by(
        func.count().desc(),
    ).limit(k)
    result = conn.execute(stmt)

    topk = {
        "values": [],
        "counts": [],
    }
    for row in result:
        k, v = row
        if k is not None:
            k = str(k)
        topk["values"].append(k)
        topk["counts"].append(v)
    return topk


def profile_histogram(
    conn: Connection,
    table: FromClause,
    column: ColumnClause,
    min: Union[int, float],
    max: Union[int, float],
    is_integer: bool,
    num_buckets: int = HISTOGRAM_NUM_BUCKET
) -> dict:
    if is_integer:
        # min=0, max=50, num_buckets=50  => interval=1, num_buckets=51
        # min=0, max=70, num_buckets=50  => interval=2, num_buckets=36
        # min=0, max=100, num_buckets=50 => interval=2, num_buckets=51
        interval = math.ceil((max - min) / num_buckets) if max > min else 1
        num_buckets = math.ceil((max - min + 1) / interval)
    else:
        interval = (max - min) / num_buckets if max > min else 1

    cases = []
    for i in range(num_buckets):
        bound = min + interval * (i + 1)
        if i != num_buckets - 1:
            cases += [(column < bound, i)]
        else:
            cases += [(column < bound + interval / 100, i)]

    cte_with_bucket = select(
        column.label("c"),
        case(*cases, else_=None).label("bucket")
    ).select_from(
        table
    ).where(
        column.isnot(None)
    ).cte()

    stmt = select(
        cte_with_bucket.c.bucket,
        func.count().label("_count")
    ).group_by(
        cte_with_bucket.c.bucket
    ).order_by(
        cte_with_bucket.c.bucket
    )

    result = conn.execute(stmt)

    counts = []
    labels = []
    bin_edges = []
    for i in range(num_buckets):
        if is_integer:
            start = min + i * interval
            end = min + (i + 1) * interval
            if interval == 1:
                label = f"{start}"
            else:
                label = f"{start} _ {end}"
        else:
            if interval >= 1:
                start = min + i * interval
                end = min + (i + 1) * interval
            else:
                start = min + i / (1 / interval)
                end = min + (i + 1) / (1 / interval)

            label = f"{format_float(start)} _ {format_float(end)}"

        labels.append(label)
        counts.append(0)
        bin_edges.append(start)
        if i == num_buckets - 1:
            bin_edges.append(end)

    for row in result:
        _bucket, v = row
        if _bucket is None:
            continue
        counts[int(_bucket)] = v
    return {
        "labels": labels,
        "counts": counts,
        "bin_edges": bin_edges,
    }


def profile_non_duplicate(
    conn: Connection,
    table: FromClause,
    column: ColumnClause,
) -> int:
    # with t as (
    #     select count(column) as c
    # from table
    # group by column
    # having c == 1
    # )
    # select
    # count(c) as non_duplicate
    # from t;
    cte = select(
        func.count(column).label("non_duplicates")
    ).select_from(
        table
    ).where(
        column.isnot(None)
    ).group_by(
        column
    ).having(
        func.count(column) == 1
    ).cte()

    stmt = select(func.count(cte.c.non_duplicates)).select_from(cte)
    non_duplicates, = conn.execute(stmt).fetchone()
    return non_duplicates



import re
from ipaddress import ip_network, ip_address

import requests
from sqlalchemy import text
from sqlalchemy.engine.url import URL
from sqlalchemy.sql import quoted_name
from sqlalchemy.sql.functions import OrderedSetAgg

from piperider_cli.error import PipeRiderConnectorError, AwsCredentialsError
from . import DataSource
from .field import TextField, ListField, PasswordField

AUTH_METHOD_PASSWORD = 'password'
AUTH_METHOD_IAM = 'iam'


class ApproximatePercentileDisc(OrderedSetAgg):
    identifier = "approximate_percentile_disc"
    inherit_cache = True
    name = quoted_name(text("APPROXIMATE PERCENTILE_DISC"), False)
    array_for_multi_clause = True


def _is_redshift_serverless(ans) -> bool:
    # Example:
    #   cluster-endpoint: redshift-cluster-1.xxx.ap-northeast-1.redshift.amazonaws.com:5439/dev
    #   serverless-endpoint: default.xxx.ap-northeast-1.redshift-serverless.amazonaws.com:5439/dev
    endpoint = ans.get('endpoint')
    return 'redshift-serverless' == endpoint.split('.')[3]


def _is_iam_auth(ans) -> bool:
    return AUTH_METHOD_IAM == ans.get('method')


def _is_password_auth(ans) -> bool:
    return AUTH_METHOD_PASSWORD == ans.get('method')


def _is_ip_address(host) -> bool:
    obj = re.search(r"^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})$", host)
    if obj is None:
        return False
    else:
        for v in obj.groups():
            if int(v) > 255:
                return False
    return True


def find_aws_region(ip):
    ip_json = requests.get('https://ip-ranges.amazonaws.com/ip-ranges.json').json()
    prefixes = ip_json['prefixes']
    my_ip = ip_address(ip)
    region = 'Unknown'
    for prefix in prefixes:
        if my_ip in ip_network(prefix['ip_prefix']):
            region = prefix['region']
            break
    return region


class RedshiftDataSource(DataSource):
    def __init__(self, name, **kwargs):
        super().__init__(name, 'redshift', **kwargs)
        self.fields = [
            TextField('endpoint', description='Redshift Endpoint'),
            ListField('method', description='Authentication Methods',
                      default=[AUTH_METHOD_PASSWORD, AUTH_METHOD_IAM]),
            TextField('user', description='Redshift User',
                      ignore=lambda answer: _is_iam_auth(answer) and _is_redshift_serverless(answer)),
            PasswordField('password', description='Redshift Password',
                          ignore=_is_iam_auth),
            TextField('iam_profile', description='Redshift IAM Role', optional=True,
                      ignore=_is_password_auth),
            TextField('dbname', description='Redshift Database Name', default='dev'),
            TextField('schema', description='Redshift Schema', default='public'),
        ]

    def validate(self):
        if self.type_name != 'redshift':
            raise ValueError('type name should be redshift')
        return self._validate_required_fields()

    def to_database_url(self, database):
        credential = self.credential
        if database is None:
            database = credential.get('dbname')
        if database is None:
            database = credential.get('database')

        host = credential.get('host') or credential.get('endpoint').split('/')[0].split(':')[0]
        try:
            port = credential.get('port') or credential.get('endpoint').split('/')[0].split(':')[1]
        except Exception:
            port = 5439

        from sqlalchemy_redshift.dialect import RedshiftDialect
        RedshiftDialect.supports_statement_cache = True

        if credential.get('method') == AUTH_METHOD_PASSWORD:
            user = credential.get('user')
            password = credential.get('password')
        else:
            import boto3
            from botocore.config import Config

            try:
                iam_profile = credential.get('iam_profile')
                aws = boto3.Session(profile_name=iam_profile)
            except Exception:
                aws = boto3

            if _is_ip_address(host):
                region = find_aws_region(host)
            else:
                region = host.split('.')[2]

            try:
                if _is_redshift_serverless(credential):
                    # Redshift Serverless
                    client = aws.client('redshift-serverless',
                                        config=Config(region_name=region))
                    workgroup = host.split('.')[0]
                    cluster_creds = client.get_credentials(dbName=database,
                                                           workgroupName=workgroup,
                                                           durationSeconds=3600)
                    user = cluster_creds['dbUser']
                    password = cluster_creds['dbPassword']
                else:
                    # Redshift Provisioned Clusters
                    user = credential.get('user')
                    cluster_id = credential.get('cluster_id') or host.split('.')[0]
                    duration_s = credential.get('iam_duration_seconds', 900)
                    auto_create = credential.get('auto_create', False)
                    db_groups = credential.get('db_groups', [])
                    client = aws.client('redshift',
                                        config=Config(region_name=region))
                    cluster_creds = client.get_cluster_credentials(DbUser=user,
                                                                   DbName=database,
                                                                   ClusterIdentifier=cluster_id,
                                                                   DurationSeconds=duration_s,
                                                                   AutoCreate=auto_create,
                                                                   DbGroups=db_groups, )
                    user = cluster_creds.get('DbUser')
                    password = cluster_creds.get('DbPassword')
                    if _is_ip_address(host):
                        host = client.describe_clusters(ClusterIdentifier=cluster_id)['Clusters'][0]['Endpoint'][
                            'Address']
            except Exception as e:
                raise AwsCredentialsError(str(e))

        url = URL.create(drivername='redshift+psycopg2',
                         host=host,
                         port=port,
                         username=user, password=password, database=database)
        return url

    def engine_args(self):
        args = dict(connect_args={'connect_timeout': 5})
        return args

    def verify_connector(self):
        try:
            import psycopg2
            import redshift_connector
            import sqlalchemy_redshift
            import boto3
            # do nothing when everything is ok
            return None
        except Exception as e:
            return PipeRiderConnectorError(str(e), 'redshift')

    def _get_display_description(self):
        cred = self.credential
        database = cred.get('dbname') if cred.get('dbname') else cred.get('database')
        return f"type={self.type_name}, dbname={database}, schema={cred.get('schema')}"

    def get_database(self):
        return self.credential.get('dbname')

    def get_schema(self):
        return self.credential.get('schema')



import json
import os
import shlex
import subprocess
import sys
from subprocess import Popen
from typing import Dict

from piperider_cli.error import RecipeException
from piperider_cli.recipes import RecipeConfiguration, select_recipe_file


def git_switch_to(branch_name):
    outs, errs, exit_code = run_external_command(f"git switch {branch_name}")
    if exit_code != 0:
        raise RecipeException(errs)


def run_external_command(command_line, env: Dict = None):
    cmd = shlex.split(command_line)
    proc = None
    try:
        proc = Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env or os.environ.copy())
        outs, errs = proc.communicate()
    except BaseException as e:
        if proc:
            proc.kill()
            outs, errs = proc.communicate()
        else:
            return None, e, 1

    if outs is not None:
        outs = outs.decode().strip()
    if errs is not None:
        errs = errs.decode().strip()
    return outs, errs, proc.returncode


def prepare_for_action(recipe_name: str = None):
    recipe_path = select_recipe_file(None if not recipe_name else recipe_name)
    if recipe_path is not None:
        cfg = RecipeConfiguration.load(recipe_path)

        if cfg.base.branch:
            git_switch_to(cfg.base.branch)
        if cfg.target.branch:
            git_switch_to(cfg.target.branch)

    ref_name = os.environ.get('GITHUB_HEAD_REF')
    print(f'switch to GITHUB_HEAD_REF: {ref_name}')
    git_switch_to(ref_name)


def make_recipe_command():
    command_builder = ["piperider compare", "--summary-file ./summary.md", "-o ./report"]
    recipe = os.environ.get("INPUT_RECIPE")
    if recipe:
        command_builder.append(f"--recipe {recipe}")

    api_token = os.environ.get("INPUT_CLOUD_API_TOKEN")
    if api_token:
        # TODO check login success or failed
        project = os.environ.get("INPUT_CLOUD_PROJECT")

        if project:
            print(f"piperider cloud login --token {api_token} --project {project} --no-interaction")
        else:
            print(f"piperider cloud login --token {api_token} --no-interaction")

        share = os.environ.get("INPUT_SHARE", "false")
        if share == "true":
            command_builder.append("--share")

        upload = os.environ.get("INPUT_UPLOAD", "false")
        if upload == "true":
            command_builder.append("--upload")

    compare_command = " ".join(command_builder)
    print(compare_command)


def create_comment(github_token: str, repo: str, issue_number: str, body: str):
    # https://docs.github.com/en/rest/issues/comments?apiVersion=2022-11-28#create-an-issue-comment
    import requests
    url = f"https://api.github.com/repos/{repo}/issues/{issue_number}/comments"
    response = requests.post(url=url, json=dict(body=body),
                             headers={"accept": "application/vnd.github+json",
                                      "X-GitHub-Api-Version": "2022-11-28",
                                      "Authorization": f"token {github_token}"})

    if response.status_code != 201:
        print("[Error] Failed to create pull-request comment", response.status_code, response.text)
        sys.exit(1)


def attach_comment():
    if not os.path.exists("./summary.md"):
        print("./summary.md not found")
        sys.exit(1)
        return

    summary = ""
    with open("./summary.md") as fh:
        summary = fh.read()
        if ":bar_chart:" not in summary:
            summary = f"""
# :bar_chart: Piperider Comparison Summary
{summary}

## :paperclip: Generated Comparison Report ZIP
Find it in the [Github Action Runs Page]({os.environ.get('GITHUB_ACTION_URL')})
"""

    with open("/github/workflow/event.json", "r") as fh:
        event = json.loads(fh.read())
        number = event['number']
        create_comment(os.environ['GITHUB_TOKEN'], os.environ['GITHUB_REPOSITORY'], number, summary)


if __name__ == '__main__':
    function = sys.argv[-1]
    try:
        if function == "prepare_for_action":
            recipe = os.environ.get("INPUT_RECIPE")
            prepare_for_action(recipe)
        elif function == "make_recipe_command":
            make_recipe_command()
        elif function == "attach_comment":
            attach_comment()
        else:
            raise Exception(f"Unknown command: {function}")
    except Exception as e:
        print(f"[Error] {e}")
        sys.exit(1)



import json
import os
import unittest
from typing import Dict
from unittest import TestCase

from packaging import version

from piperider_cli.dbt.list_task import (
    compare_models_between_manifests,
    dbt_version_obj, list_resources_from_manifest,
    ResourceSelector,
    load_manifest,
)
from piperider_cli.dbt.changeset import GraphDataChangeSet


class _BaseDbtTest(TestCase):
    def setUp(self):
        self.maxDiff = None
        self.dbt_state_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "mock_dbt_data"
        )

    def fake_data_path(self, filename: str):
        return os.path.join(self.dbt_state_dir, filename)

    def manifest_object(self, manifest_name, embedded_from_run=False):
        with open(self.fake_data_path(manifest_name)) as fh:
            data = json.loads(fh.read())
            if embedded_from_run:
                data = data.get("dbt", {}).get("manifest", {})
            return load_manifest(data)

    def run_object(self, run_name):
        with open(self.fake_data_path(run_name)) as fh:
            data = json.loads(fh.read())
            return data

    def load_json(self, filename: str) -> Dict:
        with open(self.fake_data_path(filename)) as fh:
            return json.loads(fh.read())

    def base_manifest(self):
        return self.manifest_object("jaffle_shop_base_1.3.json", True)

    def target_manifest(self):
        return self.manifest_object("jaffle_shop_target_1.3.json", True)

    def base_run(self):
        return self.run_object("jaffle_shop_base_1.3.json")

    def target_run(self):
        return self.run_object("jaffle_shop_target_1.3.json")

    def base_31587(self):
        return self.run_object("sc-31587-base.json")

    def target_31587(self):
        return self.run_object("sc-31587-input.json")

    def base_31587_with_ref(self):
        return self.run_object("sc-31587-with-ref-base.json")

    def target_31587_with_ref(self):
        return self.run_object("sc-31587-with-ref-input.json")

    def base_31587_metrics(self):
        return self.run_object("sc-31587-metrics-base.json")

    def target_31587_metrics(self):
        return self.run_object("sc-31587-metrics-input.json")

    def base_31782(self):
        return self.run_object("sc-31782-base.json")

    def target_31782(self):
        return self.run_object("sc-31782-target.json")


class TestDbtIntegration(_BaseDbtTest):
    def setUp(self):
        super().setUp()

    def test_list_all_resources(self):
        expected = [
            "metric:jaffle_shop.revenue",
            "jaffle_shop.customers",
            "jaffle_shop.orders",
            "jaffle_shop.staging.stg_customers",
            "jaffle_shop.staging.stg_orders",
            "jaffle_shop.staging.stg_payments",
            "jaffle_shop.raw_customers",
            "jaffle_shop.raw_orders",
            "jaffle_shop.raw_payments",
            "jaffle_shop.accepted_values_orders_status__placed__shipped__completed__return_pending__returned",
            "jaffle_shop.staging.accepted_values_stg_orders_status__placed__shipped__completed__return_pending__returned",
            "jaffle_shop.staging.accepted_values_stg_payments_payment_method__credit_card__coupon__bank_transfer__gift_card",
            "jaffle_shop.not_null_customers_customer_id",
            "jaffle_shop.not_null_orders_amount",
            "jaffle_shop.not_null_orders_bank_transfer_amount",
            "jaffle_shop.not_null_orders_coupon_amount",
            "jaffle_shop.not_null_orders_credit_card_amount",
            "jaffle_shop.not_null_orders_customer_id",
            "jaffle_shop.not_null_orders_gift_card_amount",
            "jaffle_shop.not_null_orders_order_id",
            "jaffle_shop.staging.not_null_stg_customers_customer_id",
            "jaffle_shop.staging.not_null_stg_orders_order_id",
            "jaffle_shop.staging.not_null_stg_payments_payment_id",
            "jaffle_shop.relationships_orders_customer_id__customer_id__ref_customers_",
            "jaffle_shop.unique_customers_customer_id",
            "jaffle_shop.unique_orders_order_id",
            "jaffle_shop.staging.unique_stg_customers_customer_id",
            "jaffle_shop.staging.unique_stg_orders_order_id",
            "jaffle_shop.staging.unique_stg_payments_payment_id",
        ]
        all_results = list_resources_from_manifest(self.base_manifest())

        self.assertListEqual(expected, all_results)

    def test_list_models(self):
        expected = [
            "jaffle_shop.customers",
            "jaffle_shop.orders",
            "jaffle_shop.staging.stg_customers",
            "jaffle_shop.staging.stg_orders",
            "jaffle_shop.staging.stg_payments",
        ]
        all_results = list_resources_from_manifest(
            self.base_manifest(), ResourceSelector().model()
        )
        self.assertListEqual(expected, all_results)

    def test_list_seeds(self):
        expected = [
            "jaffle_shop.raw_customers",
            "jaffle_shop.raw_orders",
            "jaffle_shop.raw_payments",
        ]
        all_results = list_resources_from_manifest(
            self.base_manifest(), ResourceSelector().seed()
        )
        self.assertListEqual(expected, all_results)

    def test_compare_with_manifests(self):
        without_downstream = compare_models_between_manifests(
            self.base_manifest(), self.target_manifest()
        )

        expected = ["jaffle_shop.customers", "jaffle_shop.orders"]
        self.assertListEqual(without_downstream, expected)

        with_downstream = compare_models_between_manifests(
            self.base_manifest(), self.target_manifest(), True
        )

        expected = ["jaffle_shop.customers", "jaffle_shop.orders"]
        self.assertListEqual(with_downstream, expected)

    def test_list_explicit_changes(self):
        c = GraphDataChangeSet(self.base_run(), self.target_run())

        expected = ['model.jaffle_shop.customers', 'model.jaffle_shop.orders',
                    'test.jaffle_shop.accepted_values_orders_status'
                    '__placed__shipped__completed__return_pending__returned.be6b5b5ec3',
                    'test.jaffle_shop.not_null_customers_customer_id.5c9bf9911d',
                    'test.jaffle_shop.not_null_orders_amount.106140f9fd',
                    'test.jaffle_shop.not_null_orders_bank_transfer_amount.7743500c49',
                    'test.jaffle_shop.not_null_orders_coupon_amount.ab90c90625',
                    'test.jaffle_shop.not_null_orders_credit_card_amount.d3ca593b59',
                    'test.jaffle_shop.not_null_orders_customer_id.c5f02694af',
                    'test.jaffle_shop.not_null_orders_gift_card_amount.413a0d2d7a',
                    'test.jaffle_shop.not_null_orders_order_id.cf6c17daed',
                    'test.jaffle_shop.relationships_orders_customer_id__customer_id__ref_customers_.c6ec7f58f2',
                    'test.jaffle_shop.unique_customers_customer_id.c5af1ff4b1',
                    'test.jaffle_shop.unique_orders_order_id.fed79b3a6e']
        changes = c.list_explicit_changes()
        self.assertListEqual(changes, expected)

        print(c.list_implicit_changes())

    @unittest.skipIf(dbt_version_obj() < version.parse('1.4'),
                     'this only works after manifests generated after the v1.4')
    def test_list_explicit_changes_without_ref_ids(self):
        c = GraphDataChangeSet(self.base_31587(), self.target_31587())

        expected = ['model.jaffle_shop.orders',
                    'test.jaffle_shop.accepted_values_orders_status__placed__shipped__completed__return_pending__returned.be6b5b5ec3',
                    'test.jaffle_shop.not_null_orders_amount.106140f9fd',
                    'test.jaffle_shop.not_null_orders_bank_transfer_amount.7743500c49',
                    'test.jaffle_shop.not_null_orders_coupon_amount.ab90c90625',
                    'test.jaffle_shop.not_null_orders_credit_card_amount.d3ca593b59',
                    'test.jaffle_shop.not_null_orders_customer_id.c5f02694af',
                    'test.jaffle_shop.not_null_orders_gift_card_amount.413a0d2d7a',
                    'test.jaffle_shop.not_null_orders_order_id.cf6c17daed',
                    'test.jaffle_shop.unique_orders_order_id.fed79b3a6e']
        changes = c.list_explicit_changes()
        self.assertListEqual(changes, expected)

        expected_implicit = ['metric.jaffle_shop.average_order_amount']
        self.assertListEqual(c.list_implicit_changes(), expected_implicit)

    @unittest.skipIf(dbt_version_obj() < version.parse('1.4'),
                     'this only works after manifests generated after the v1.4')
    def test_list_explicit_changes_with_ref_ids(self):
        c = GraphDataChangeSet(self.base_31587_with_ref(), self.target_31587_with_ref())

        expected = ['model.jaffle_shop.orders',
                    'test.jaffle_shop.accepted_values_orders_status__placed__shipped__completed__return_pending__returned.be6b5b5ec3',
                    'test.jaffle_shop.not_null_orders_amount.106140f9fd',
                    'test.jaffle_shop.not_null_orders_bank_transfer_amount.7743500c49',
                    'test.jaffle_shop.not_null_orders_coupon_amount.ab90c90625',
                    'test.jaffle_shop.not_null_orders_credit_card_amount.d3ca593b59',
                    'test.jaffle_shop.not_null_orders_customer_id.c5f02694af',
                    'test.jaffle_shop.not_null_orders_gift_card_amount.413a0d2d7a',
                    'test.jaffle_shop.not_null_orders_order_id.cf6c17daed',
                    'test.jaffle_shop.unique_orders_order_id.fed79b3a6e']
        changes = c.list_explicit_changes()
        self.assertListEqual(changes, expected)

        expected_implicit = ['metric.jaffle_shop.average_order_amount']
        self.assertListEqual(c.list_implicit_changes(), expected_implicit)

    @unittest.skipIf(dbt_version_obj() < version.parse('1.4'),
                     'this only works after manifests generated after the v1.4')
    def test_list_changes_metrics_case(self):
        c = GraphDataChangeSet(self.base_31587_metrics(), self.target_31587_metrics())
        expected = ['model.jaffle_shop.orders',
                    'test.jaffle_shop.accepted_values_orders_status__placed__shipped__completed__return_pending__returned.be6b5b5ec3',
                    'test.jaffle_shop.not_null_orders_amount.106140f9fd',
                    'test.jaffle_shop.not_null_orders_bank_transfer_amount.7743500c49',
                    'test.jaffle_shop.not_null_orders_coupon_amount.ab90c90625',
                    'test.jaffle_shop.not_null_orders_credit_card_amount.d3ca593b59',
                    'test.jaffle_shop.not_null_orders_customer_id.c5f02694af',
                    'test.jaffle_shop.not_null_orders_gift_card_amount.413a0d2d7a',
                    'test.jaffle_shop.not_null_orders_order_id.cf6c17daed',
                    'test.jaffle_shop.unique_orders_order_id.fed79b3a6e']
        changes = c.list_explicit_changes()
        self.assertListEqual(changes, expected)

        changes = c.list_implicit_changes()
        self.assertListEqual(c.list_implicit_changes(), ['metric.jaffle_shop.average_order_amount'])

    @unittest.skipIf(dbt_version_obj() < version.parse('1.4'),
                     'this only works after manifests generated after the v1.4')
    def test_list_large_seeds_case(self):
        try:
            c = GraphDataChangeSet(self.base_31782(), self.target_31782())
        except Exception as e:
            self.fail("Unexpected Exception")

        expected = ['metric.jaffle_shop.expenses',
                    'model.jaffle_shop.stg_payments',
                    'test.jaffle_shop.accepted_values_stg_payments_payment_method__credit_card__coupon__bank_transfer__gift_card.3c3820f278',
                    'test.jaffle_shop.not_null_stg_payments_payment_id.c19cc50075',
                    'test.jaffle_shop.unique_stg_payments_payment_id.3744510712']

        changes = c.list_explicit_changes()
        self.assertListEqual(changes, expected)
        changes = c.list_implicit_changes()
        self.assertListEqual(changes, ['metric.jaffle_shop.profit'])



import os
import re
import sys
import threading
from abc import ABCMeta, abstractmethod
from typing import List, Dict, Callable

import inquirer
import readchar
from rich.console import Console
from rich.prompt import Prompt
from sqlalchemy import create_engine, select, text
from sqlalchemy.pool import SingletonThreadPool

import piperider_cli.hack.datasource_inquirer_prompt as datasource_prompt
from .field import DataSourceField


def _should_use_fancy_user_input() -> bool:
    env_flag = os.environ.get('PIPERIDER_FANCY_USER_INPUT', 'true').lower() == 'true'
    is_a_tty = sys.stdin.isatty() and sys.stdout.isatty()
    if env_flag is False:
        return False
    if is_a_tty is False:
        return False
    if sys.platform == "darwin" or sys.platform == "linux":
        return True
    else:
        return False


DATASOURCE_NAME_REGEX = r'^[a-zA-Z0-9_-]+$'
FANCY_USER_INPUT = _should_use_fancy_user_input()


class DataSource(metaclass=ABCMeta):
    def __init__(self, name, type_name: str, credential=None, **kwargs):
        self.name = name
        self.type_name = type_name
        self.args = kwargs
        self.fields: List[DataSourceField] = []
        self.credential: Dict = credential or {}
        self.credential_source = 'credentials'
        self._cached_engine = {}
        self._cached_lock = threading.Lock()

    def _validate_required_fields(self):
        reasons = []
        # check required fields
        for f in self.fields:
            if isinstance(f.ignore, Callable) and f.ignore(self.credential):
                continue
            if f.name not in self.credential and f.optional is False and not f.name.startswith("_"):
                reasons.append(f"{f.name} is required")

        # check if got duplicate keys: pass & password
        if self.credential.get('pass') and self.credential.get('password'):
            reasons.append('Got duplicate keys: (pass) all map to "password"')

        return reasons == [], reasons

    @abstractmethod
    def validate(self):
        """
        validate type name and required fields.

        Returns True if everything is fine, False and reasons otherwise.

        :return: bool, []
        """
        raise NotImplementedError

    @abstractmethod
    def to_database_url(self, database):
        """
        build a database url for sqlalchemy create_engine method
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def verify_connector(self):
        raise NotImplementedError

    def verify_connection(self):
        engine = self.get_engine_by_database()
        with engine.connect() as conn:
            stmt = select(text('1 as id'))
            conn.execute(stmt)
        return

    def create_engine(self, database=None):
        return create_engine(self.to_database_url(database=database), **self.engine_args())

    def get_engine_by_database(self, database=None):
        engine = self._cached_engine.get(database)
        if engine is None:
            with self._cached_lock:
                engine = self._cached_engine.get(database)
                if engine is None:
                    engine = self.create_engine(database)
                    self._cached_engine[database] = engine

        return engine

    @property
    def threads(self):
        if self.credential.get('threads'):
            return self.credential.get('threads')
        elif not isinstance(self.get_engine_by_database().pool, SingletonThreadPool):
            return 5
        else:
            return 1

    def engine_args(self):
        return dict()

    def show_installation_information(self):
        from rich.markup import escape
        err = self.verify_connector()
        if err:
            console = Console()
            console.print(f'\n{escape(err.hint)}\n')
            return False
        return True

    def ask_credential(self):
        """
        ask for user filling all fields.
        """
        if FANCY_USER_INPUT:
            return self.ask_credential_by_inquirer()
        else:
            return self.ask_credential_by_rich()

    def ask_credential_by_inquirer(self):
        if sys.platform == "darwin" or sys.platform == "linux":
            # change readchar key backspace
            readchar.key.BACKSPACE = '\x7F'

        def next_question(fields):
            for f in self.fields:
                yield f.question()

        answers = datasource_prompt.prompt(next_question(self.fields), raise_keyboard_interrupt=True)
        if answers:
            for f in self.fields:
                if answers[f.name]:
                    self.credential[f.name] = answers[f.name].strip()
            self.credential = {k: v for (k, v) in self.credential.items() if not k.startswith('_')}
        return self.credential

    def ask_credential_by_rich(self):
        for f in self.fields:
            if hasattr(f, 'callback') and getattr(f, 'callback') is not None:
                answer = f.callback(self.credential)
            else:
                answer = f.question_by_rich(self.credential)
            if answer:
                self.credential[f.name] = answer.strip()

        self.credential = {k: v for (k, v) in self.credential.items() if not k.startswith('_')}
        return self.credential

    @staticmethod
    def ask(exist_datasource: List[str] = None):
        if FANCY_USER_INPUT:
            return DataSource.ask_by_inquirer(exist_datasource=exist_datasource)
        else:
            return DataSource.ask_by_rich(exist_datasource=exist_datasource)

    @staticmethod
    def ask_by_rich(exist_datasource: List[str] = None):
        if exist_datasource is None:
            exist_datasource = []
        console = Console()
        source_choices = [(k, v) for k, v in DATASOURCE_PROVIDERS.items()]

        while True:
            datasource_name = Prompt.ask(
                "[[yellow]?[/yellow]] What is your data source name? (alphanumeric and underscore are allowed)")
            if re.match(DATASOURCE_NAME_REGEX, datasource_name) and datasource_name not in exist_datasource:
                break
            else:
                console.print('    [[red]Error[/red]] Input is not a valid datasource name. Please try again.')

        console.print('[[yellow]?[/yellow]] Which data source would you like to connect to?')
        for i, (k, v) in enumerate(source_choices):
            console.print(f'  [green]{i + 1}[/green]: {k}')

        while True:
            try:
                type_idx = Prompt.ask('[[yellow]?[/yellow]] Select a number: ')
                type_idx = int(type_idx)
            except EOFError as e:
                raise e
            except Exception:
                type_idx = 0

            if type_idx > len(source_choices) or type_idx < 1:
                console.print('    [[red]Error[/red]] Input is not a valid index value. Please try again.')
            else:
                cls = source_choices[type_idx - 1][1]
                break

        name = datasource_name
        return cls, name

    @staticmethod
    def ask_by_inquirer(exist_datasource: List[str] = None):
        if exist_datasource is None:
            exist_datasource = []
        source_choices = [(k, v) for k, v in DATASOURCE_PROVIDERS.items()]

        if sys.platform == "darwin" or sys.platform == "linux":
            # change readchar key backspace
            readchar.key.BACKSPACE = '\x7F'

        questions = [
            inquirer.Text('datasource_name',
                          message='What is your data source name? (alphanumeric and underscore are allowed)',
                          validate=lambda ans, x: re.match(DATASOURCE_NAME_REGEX,
                                                           x) is not None and x not in exist_datasource),
            inquirer.List('type',
                          message='Which data source would you like to connect to?',
                          choices=source_choices,
                          ),
        ]
        answers = inquirer.prompt(questions, raise_keyboard_interrupt=True)
        name = answers['datasource_name'].strip()
        cls = answers['type']
        return cls, name

    @abstractmethod
    def _get_display_description(self):
        """
        get the description of the datasource to display.

        Returns description

        :return: str
        """
        raise NotImplementedError

    def get_display_description(self):
        # if dbt
        dbt = self.args.get('dbt')
        if dbt:
            return f"profile={dbt.get('profile')}, target={dbt.get('target')}"
        return self._get_display_description()

    @abstractmethod
    def get_database(self):
        raise NotImplementedError

    @abstractmethod
    def get_schema(self):
        raise NotImplementedError


def _list_datasource_providers():
    from .snowflake import SnowflakeDataSource
    from .postgres import PostgresDataSource
    from .sqlite import SqliteDataSource
    from .bigquery import BigQueryDataSource
    from .redshift import RedshiftDataSource
    from .survey import UserSurveyMockDataSource
    from .duckdb import DuckDBDataSource, CsvDataSource, ParquetDataSource
    from .athena import AthenaDataSource
    return {
        'snowflake': SnowflakeDataSource,
        'bigquery': BigQueryDataSource,
        'redshift': RedshiftDataSource,
        'athena': AthenaDataSource,
        'postgres': PostgresDataSource,
        'sqlite': SqliteDataSource,
        'duckdb': DuckDBDataSource,
        'csv': CsvDataSource,
        'parquet': ParquetDataSource,
        'tell us what type of datasource you want': UserSurveyMockDataSource,
    }


DATASOURCE_PROVIDERS = _list_datasource_providers()



import abc
import os
import re
import shlex
import subprocess
import tarfile
import tempfile
from subprocess import Popen
from typing import Dict, Tuple

from rich.console import Console

from piperider_cli.configuration import FileSystem
from piperider_cli.dbt.list_task import list_resources_from_manifest
from piperider_cli.error import PipeRiderError, RecipeException


class AbstractRecipeUtils(metaclass=abc.ABCMeta):
    def __init__(self, console: Console):
        self.console = console

    @abc.abstractmethod
    def execute_command_in_silent(
        self, command_line, env: Dict = None
    ) -> Tuple[str, str, int]:
        """
        Execute command without showing outputs

        the outputs will be collected to the method returns

        :returns: stdout, stderr, exit_code
        """

    @abc.abstractmethod
    def execute_command_with_showing_output(
        self, command_line, env: Dict = None
    ) -> int:
        """
        Execute command and showing outputs

        the outputs will be showing to the terminal

        :returns: exit_code
        """

    @staticmethod
    def dryrun_ignored_execute_command(command_line, env: Dict = None):
        cmd = shlex.split(command_line)
        proc = None
        try:
            proc = Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=env or os.environ.copy(),
                cwd=FileSystem.WORKING_DIRECTORY,
            )
            outs, errs = proc.communicate()
        except BaseException as e:
            if proc:
                proc.kill()
                outs, errs = proc.communicate()
            else:
                return None, e, 1

        if outs is not None:
            outs = outs.decode().strip()
        if errs is not None:
            errs = errs.decode().strip()
        return outs, errs, proc.returncode

    @staticmethod
    def dryrun_ignored_execute_command_no_outputs(command_line, env: Dict = None):
        cmd = shlex.split(command_line)
        proc = None

        try:
            proc = Popen(cmd, env=env or os.environ.copy(), cwd=FileSystem.WORKING_DIRECTORY)
            proc.communicate()
        except BaseException:
            if proc:
                proc.kill()
                proc.communicate()
            else:
                return 1

        return proc.returncode

    def git_branch(self):
        outs, errs, exit_code = self.dryrun_ignored_execute_command(
            "git rev-parse --abbrev-ref HEAD"
        )
        if exit_code != 0:
            return None

        return outs

    def git_checkout_to(self, commit_or_branch):
        outs, errs, exit_code = self.execute_command_in_silent(
            f"git checkout {commit_or_branch}"
        )
        if exit_code != 0:
            raise RecipeException(errs)

    def git_merge_base(self, a: str, b: str):
        outs, errs, exit_code = self.dryrun_ignored_execute_command(
            f"git merge-base {a} {b}"
        )
        if exit_code != 0:
            ex = RecipeException(errs)
            if outs == "" and errs == "":
                ex.message = "Empty result from git merge-base"
                ex.hint = 'Please try "git fetch --unshallow" first'
            elif re.match(r"fatal: Not a valid object name (.*)", errs):
                matched = re.match(r"fatal: Not a valid object name (.*)", errs)
                ex.message = f"Invalid git branch: {matched.group(1)}"
                ex.hint = (
                    f"Please check is the branch name '{matched.group(1)}' correct?"
                )
            raise ex
        return outs

    def git_archive(self, commit_or_branch):
        def untar(file_path, extract_dir):
            with tarfile.open(file_path, 'r') as tar:
                tar.extractall(extract_dir)

        tmpdirname = tempfile.mkdtemp()
        tar = os.path.join(tmpdirname, f'{commit_or_branch}.tar')
        outs, errs, exit_code = self.execute_command_in_silent(
            f"git archive --format=tar --output={tar} {commit_or_branch}"
        )
        if exit_code != 0:
            raise RecipeException(errs)

        project_dir = os.path.join(tmpdirname, commit_or_branch)
        untar(tar, project_dir)

        return project_dir

    def ensure_git_ready(self):
        outs, errs, exit_code = self.dryrun_ignored_execute_command("git --version")
        if exit_code != 0:
            raise PipeRiderError("git is not installed", hint="Please install it first")

        if "version" not in outs:
            raise RecipeException("Unknown response from git --version")

        outs, errs, exit_code = self.dryrun_ignored_execute_command(
            "git status --porcelain"
        )
        if exit_code != 0 and "not a git repository" in errs:
            raise RecipeException("The working directory is not a git repository.")

    def check_dbt_command(self):
        outs, errs, exit_code = self.dryrun_ignored_execute_command("dbt --version")
        if exit_code != 0:
            raise PipeRiderError("dbt is not installed", hint="Please install it first")

    def list_dbt_resources(self, manifest, select=None, state=None):
        return list_resources_from_manifest(manifest, select=select, state=state)


class RecipeUtils(AbstractRecipeUtils):
    def execute_command_with_showing_output(self, command_line, env: Dict = None):
        return self.dryrun_ignored_execute_command_no_outputs(command_line, env)

    def execute_command_in_silent(self, command_line, env: Dict = None):
        outs, errs, exit_code = self.dryrun_ignored_execute_command(command_line, env)
        return outs, errs, exit_code


class DryRunRecipeUtils(AbstractRecipeUtils):
    def execute_command_in_silent(self, command_line, env: Dict = None):
        self.console.print(f"[green]:external-command:>[/green] {command_line}")
        return "", "", 0

    def execute_command_with_showing_output(self, command_line, env: Dict = None):
        self.console.print(f"[green]:external-command:>[/green] {command_line}")
        return 0

    def git_archive(self, commit_or_branch):
        cmd = f"git archive --format=tar --output=/path/to/tmp/{commit_or_branch}.tar {commit_or_branch}"
        self.console.print(f"[green]:external-command:>[/green] [default]{cmd}[/default]")
        return '/path/to/tmp'

    def list_dbt_resources(self, manifest, select=None, state=None):
        state_msg = ''
        if state:
            state_msg = f'state: {state}'
        self.console.print(f"[green]:dbt-resources:>[/green] with select: {''.join(select)} {state_msg}")
        return ['dbt_model1', 'dbt_model2', 'dbt_model3']

    def remove_dir(self, path):
        self.console.print(f"[green]:remove-dir:>[/green] {path}")


class InteractiveStopException(Exception):
    def __init__(self):
        pass


class InteractiveRecipeDecorator(AbstractRecipeUtils):
    def __init__(self, utils: AbstractRecipeUtils):
        self.decoratee = utils

    def execute_command_in_silent(
        self, command_line, env: Dict = None
    ) -> Tuple[str, str, int]:
        # ask user continue
        if self.should_continue(command_line):
            outs, errs, exit_code = self.decoratee.execute_command_in_silent(
                command_line, env
            )
            return outs, errs, exit_code
        raise InteractiveStopException()

    def execute_command_with_showing_output(
        self, command_line, env: Dict = None
    ) -> int:
        # ask user continue
        if self.should_continue(command_line):
            return self.decoratee.execute_command_with_showing_output(command_line, env)
        raise InteractiveStopException()

    @staticmethod
    def should_continue(command_line: str):
        try:
            return InteractiveRecipeDecorator.ask_for_confirmation(command_line)
        except KeyboardInterrupt:
            raise InteractiveStopException()

    @staticmethod
    def ask_for_confirmation(command_line: str):
        from piperider_cli.datasource import FANCY_USER_INPUT

        if FANCY_USER_INPUT:
            import inquirer
            return inquirer.confirm(f"Execute '{command_line}'", default=True)
        else:
            from rich.prompt import Confirm
            return Confirm.ask(f"Execute '{command_line}'")

    def git_archive(self, commit_or_branch) -> str:
        # ask user continue
        cmd = f"git archive --format=tar --output=/path/to/tmp/{commit_or_branch}.tar {commit_or_branch}"
        if self.should_continue(cmd):
            return self.decoratee.git_archive(commit_or_branch)
        raise InteractiveStopException()


if __name__ == "__main__":
    utils = RecipeUtils(Console())
    utils = DryRunRecipeUtils(Console())
    utils.ensure_git_ready()
    utils.git_checkout_to("abc")



import ast
import glob
import os
from _ast import AST, FunctionDef, Attribute
from typing import Any
from unittest import TestCase


class TrackCommandSetupTests(TestCase):

    def test_command_options(self):
        """
        find all decorators in the cli, like this
        @cli.command(short_help='Show version information.', cls=TrackCommand)

        to check "cls=TrackCommand" has set
        """

        import piperider_cli.cli as cli

        root = ast.parse(open(cli.__file__).read(), cli.__file__)

        results = []

        class V(ast.NodeVisitor):
            def visit_FunctionDef(self, node: FunctionDef) -> Any:
                if not node.decorator_list:
                    return

                for d in node.decorator_list:

                    # check command decorator: @*.command( ...[keywords]...)
                    is_command_decorator = hasattr(d, 'func') \
                                           and isinstance(d.func, Attribute) and d.func.attr == 'command'
                    if not is_command_decorator:
                        continue

                    has_track_setup = False

                    # check keywords has "cls"
                    for k in d.keywords:
                        if k.arg == 'cls' and k.value.id == 'TrackCommand':
                            has_track_setup = True

                    results.append((d, has_track_setup))

        # check the @cli.command
        V().visit(root)

        require_actions = [(node, result) for node, result in results if result != True]
        if require_actions:
            print(r'Please add "cls=TrackCommand" to @cli.command')
            for node, result in require_actions:
                print(f"line {node.lineno} => {ast.unparse(node)}")

            self.assert_(False, r'Please add "cls=TrackCommand" to @cli.command')



import json
import os

import pytest
from dbt.contracts.graph.manifest import WritableManifest

from piperider_cli.dbt import disable_dbt_compile_stats
from piperider_cli.dbt.list_task import dbt_version_obj, load_manifest
from packaging import version

v = dbt_version_obj()


def _load_manifest(file_name):
    file_path = os.path.join(os.path.dirname(__file__), 'mock_dbt_data', 'compatible', file_name)
    with open(file_path, 'r') as file:
        content = file.read()
    return json.loads(content)


@pytest.fixture()
def manifest_from_1_3():
    return _load_manifest('dbt-postgres-1.3.4-manifest.json')


@pytest.fixture()
def manifest_from_1_4():
    return _load_manifest('dbt-duckdb-1.4.2-manifest.json')


@pytest.fixture()
def manifest_from_1_5():
    return _load_manifest('dbt-duckdb-1.5.1-manifest.json')


def test_load_manifest_v7(manifest_from_1_3):
    assert isinstance(load_manifest(manifest_from_1_3), WritableManifest)


@pytest.mark.skipif(v < version.parse('v1.4'), reason='skip manifest test before dbt-core 1.4')
def test_load_manifest_v8(manifest_from_1_4):
    assert isinstance(load_manifest(manifest_from_1_4), WritableManifest)


@pytest.mark.skipif(v < version.parse('v1.5'), reason='skip manifest test before dbt-core 1.5')
def test_load_manifest_v9(manifest_from_1_5):
    assert isinstance(load_manifest(manifest_from_1_5), WritableManifest)


def test_log_functions():
    with disable_dbt_compile_stats():
        # make sure this not breaking in all dbt versions
        pass



import os
import sys
from abc import ABCMeta
from dataclasses import dataclass
from typing import Any, Dict, List

import jsonschema
from jsonschema.exceptions import ValidationError
from rich.console import Console
from ruamel import yaml
from ruamel.yaml import CommentedSeq

import piperider_cli.dbtutil as dbtutil
from piperider_cli import get_run_json_path, load_jinja_template, load_json
from piperider_cli.configuration import Configuration, FileSystem
from piperider_cli.dbt.list_task import load_manifest, load_full_manifest
from piperider_cli.error import RecipeConfigException
from piperider_cli.recipes.utils import InteractiveStopException

PIPERIDER_RECIPES_SCHEMA_PATH = os.path.join(os.path.dirname(__file__), 'recipe_schema.json')

console = Console()


class RecipeEnv:

    def __init__(self):
        self.name: str = None
        self.value: str = None


class AbstractRecipeField(metaclass=ABCMeta):
    environments: Dict[str, str] = {}
    commands: List[str] = []

    def __init__(self, content: dict = None):
        if content is None:
            return

        self.environments: Dict[str, str] = {e['name']: e['value'] for e in content.get('env', [])}
        self.commands: List[str] = content.get('commands', [])

        if content.get('command'):
            data = content.get('command')
            if isinstance(data, CommentedSeq):
                for c in data:
                    self.commands.append(c)
            else:
                self.commands.append(data)

    def __dict__(self):
        d = dict()
        if self.environments:
            d['env'] = [{'name': k, 'value': v} for k, v in self.environments.items()]
        if len(self.commands) == 1:
            d['command'] = self.commands[0]
        elif self.commands:
            d['commands'] = self.commands

        return d

    def envs(self):
        merged_envs = os.environ.copy()
        for k, v in self.environments.items():
            merged_envs[k] = v
        return merged_envs


class RecipeDbtField(AbstractRecipeField):
    pass


class RecipePiperiderField(AbstractRecipeField):
    pass


class RecipeCloudField(AbstractRecipeField):
    datasource: str = None
    report_id: int = None

    def __init__(self, content: dict = None):
        if content is None:
            return

        self.environments = {e['name']: e['value'] for e in content.get('env', [])}
        self.datasource = content.get('datasource')
        self.report_id = int(content.get('report_id'))

    def __dict__(self):
        return {
            'env': [{'name': k, 'value': v} for k, v in self.environments.items()],
            'datasource': self.datasource,
            'report_id': self.report_id
        }


class RecipeModel:
    branch: str = None
    dbt: RecipeDbtField = None
    piperider: RecipePiperiderField = None
    cloud: RecipeCloudField = None
    tmp_dir_path: str = None
    state_path: str = None

    def __init__(self, content: dict = None):
        if content is None:
            return

        # git branch name
        self.branch: str = content.get('branch')
        if content.get('file') is not None:
            self.file: str = content.get('file')
        self.dbt: RecipeDbtField = RecipeDbtField(content.get('dbt'))
        self.piperider: RecipePiperiderField = RecipePiperiderField(content.get('piperider'))
        self.cloud: RecipeCloudField = RecipeCloudField(content.get('cloud'))

    def __dict__(self):
        d = dict()
        if self.branch:
            d['branch'] = self.branch
        if self.is_file_specified():
            d['file'] = self.file
        if self.dbt:
            d['dbt'] = self.dbt.__dict__()
        if self.piperider:
            d['piperider'] = self.piperider.__dict__()
        if self.cloud:
            d['cloud'] = self.cloud.__dict__()
        return d

    def is_file_specified(self):
        return hasattr(self, 'file')

    def is_branch_specified(self):
        return self.branch is not None

    def is_piperider_commands_specified(self):
        return len(self.piperider.commands) > 0

    def validate_recipe(self):
        # check conflict
        if (self.is_branch_specified() or self.is_piperider_commands_specified()) and self.is_file_specified():
            raise RecipeConfigException(
                message="Both 'file' and 'branch/piperider commands' are specified.",
                hint="Please modify the recipe file to use either 'file' or 'branch/piperider commands'.")

        # check no action
        if not self.is_file_specified() and not self.is_piperider_commands_specified():
            raise RecipeConfigException(
                message="No 'file' and 'piperider commands' are given.",
                hint="Please modify the recipe file to use either 'file' or 'piperider commands.'")

        # check file existence
        if self.is_file_specified():
            if not os.path.isfile(self.file):
                raise RecipeConfigException(
                    message=f"File '{self.file}' does not exist.",
                    hint="Please modify the recipe file to use the correct file path.")

    def get_run_report(self):
        if not self.is_file_specified():
            filesystem = Configuration.instance().activate_report_directory()
            return get_run_json_path(filesystem.get_output_dir())
        return self.file


class RecipeConfiguration:
    def __init__(self, base: RecipeModel, target: RecipeModel):
        self.base: RecipeModel = base
        self.target: RecipeModel = target

    def __dict__(self):
        return {
            'base': self.base.__dict__(),
            'target': self.target.__dict__()
        }

    def dump(self):
        payload = self.__dict__()
        RecipeConfiguration.validate(payload)
        return yaml.round_trip_dump(payload)

    @staticmethod
    def validate(content: dict = None):
        schema = load_json(PIPERIDER_RECIPES_SCHEMA_PATH)
        jsonschema.validate(content, schema)

    @classmethod
    def load(cls, path: str) -> 'RecipeConfiguration':
        template = load_jinja_template(path)
        try:
            yml = yaml.YAML()
            yml.allow_duplicate_keys = True
            content = yml.load(template.render())
        except Exception:
            raise

        if content is None:
            raise Exception("Recipe content is empty")

        cls.validate(content)

        base = RecipeModel(content['base'])
        target = RecipeModel(content['target'])

        base.validate_recipe()
        target.validate_recipe()

        return cls(
            base=base,
            target=target
        )


def verify_git_dependencies(cfg: RecipeConfiguration):
    if cfg.base.branch is None and cfg.target.branch is None:
        # nobody set the git branch, skip the verification
        return

    tool().ensure_git_ready()


def verify_dbt_dependencies(cfg: RecipeConfiguration):
    if len(cfg.base.dbt.commands) == 0 and len(cfg.target.dbt.commands) == 0:
        # nobody set the dbt configurations
        return

    def test_dbt_by_deps():
        try:
            import dbt
            dbt.__path__
            return True
        except Exception:
            return False

    if test_dbt_by_deps():
        return

    # check the dbt by executing it
    tool().check_dbt_command()


def update_select_with_modified(select: tuple = None, modified: bool = False):
    if modified is False:
        return select

    if len(select) == 0:
        return ('state:modified+',)

    if any('state:modified' in item for item in select) is True:
        return select

    select_list = list(select)
    select_list[0] = select_list[0] + ',state:modified+'
    return tuple(select_list)


def prepare_dbt_resources_candidate(cfg: RecipeConfiguration, select: tuple = None, modified: bool = False):
    config = Configuration.instance()
    if not select:
        select = (f'tag:{config.dbt.get("tag")}',) if config.dbt.get('tag') else ()
    select = update_select_with_modified(select, modified)

    if any('state:' in item for item in select) is True:
        execute_dbt_compile_archive(cfg.base)

    dbt_project = dbtutil.load_dbt_project(config.dbt.get('projectDir'))
    target_path = dbt_project.get('target-path') if dbt_project.get('target-path') else 'target'
    state = cfg.base.state_path if cfg.base.state_path else None
    if state:
        console.print(f"Run: \[dbt list] select option '{' '.join(select)}' with state")
        manifest = load_full_manifest(target_path)
    else:
        console.print(f"Run: \[dbt list] select option '{' '.join(select)}'")
        manifest = load_manifest(dbtutil.get_dbt_manifest(target_path))
    console.print()
    return tool().list_dbt_resources(manifest, select=select, state=state), state


def execute_recipe(model: RecipeModel, debug=False, recipe_type='base'):
    """
    We execute a recipe in the following steps:
    1. run dbt commands
    2. run piperider commands
    """

    if model.is_file_specified():
        console.print(f"Select {recipe_type} report: \[{model.file}]")
        return

    # model.dbt.commands
    for cmd in model.dbt.commands or []:
        console.print(f"Run: \[{cmd}]")
        exit_code = tool().execute_command_with_showing_output(cmd, model.dbt.envs())
        if debug:
            console.print(f"Exit code: {exit_code}")
        if exit_code != 0:
            console.print(
                f"[bold yellow]Warning: [/bold yellow] Recipe dbt command failed: '{cmd}' with exit code: {exit_code}")
            sys.exit(exit_code)
        console.print()

    # model.piperider.commands
    for cmd in model.piperider.commands or []:
        console.print(f"Run: \[{cmd}]")
        exit_code = tool().execute_command_with_showing_output(cmd, model.piperider.envs())
        if debug:
            console.print(f"Exit code: {exit_code}")
        if exit_code != 0:
            console.print(
                f"[bold yellow]Warning: [/bold yellow] Recipe piperider command failed: '{cmd}' with exit code: {exit_code}")
            sys.exit(exit_code)
        console.print()

    # if recipe_type == 'target':
    #     config = Configuration.instance()
    #     fqn_list = dbtutil.get_fqn_list_by_tag(config.dbt.get('tag'), config.dbt.get('projectDir'))
    #     model.piperider.environments['PIPERIDER_DBT_RESOURCES'] = '\n'.join(fqn_list)


def execute_dbt_compile_archive(model: RecipeModel, debug=False):
    if not model.branch:
        raise RecipeConfigException("Branch is not specified")

    console.print("Run: \[dbt compile]")
    if model.tmp_dir_path is None:
        model.tmp_dir_path = tool().git_archive(model.branch)
        model.state_path = os.path.join(model.tmp_dir_path, 'state')

    project_dir = model.tmp_dir_path
    target_path = model.state_path
    cmd = f'dbt compile --project-dir {project_dir} --target-path {target_path}'
    exit_code = tool().execute_command_with_showing_output(cmd, model.dbt.envs())
    if exit_code != 0:
        console.print(
            f"[bold yellow]Warning: [/bold yellow] Dbt command failed: '{cmd}' with exit code: {exit_code}")
        sys.exit(exit_code)
    console.print()
    pass


def execute_recipe_archive(model: RecipeModel, debug=False, recipe_type='base'):
    """
    We execute a recipe in the following steps:
    1. export the repo with specified commit or branch if needed
    2. run dbt commands
    3. run piperider commands
    """

    if model.is_file_specified():
        console.print(f"Select {recipe_type} report: \[{model.file}]")
        return

    if model.branch:
        console.print("Run: \[git archive]")
        if model.tmp_dir_path is None:
            model.tmp_dir_path = tool().git_archive(model.branch)
        console.print()

    # model.dbt.commands
    for cmd in model.dbt.commands or []:
        console.print(f"Run: \[{cmd}]")
        # TODO: handle existing flags in command from recipe
        cmd = f'{cmd} --project-dir {model.tmp_dir_path}' if model.tmp_dir_path else cmd
        exit_code = tool().execute_command_with_showing_output(cmd, model.dbt.envs())
        if debug:
            console.print(f"Exit code: {exit_code}")
        if exit_code != 0:
            console.print(
                f"[bold yellow]Warning: [/bold yellow] Recipe dbt command failed: '{cmd}' with exit code: {exit_code}")
            sys.exit(exit_code)
        console.print()

    # model.piperider.commands
    for cmd in model.piperider.commands or []:
        console.print(f"Run: \[{cmd}]")
        cmd = f'{cmd} --dbt-project-dir {model.tmp_dir_path} --dbt-target-path {model.tmp_dir_path}/target' if model.tmp_dir_path else cmd
        exit_code = tool().execute_command_with_showing_output(cmd, model.piperider.envs())
        if debug:
            console.print(f"Exit code: {exit_code}")
        if exit_code != 0:
            console.print(
                f"[bold yellow]Warning: [/bold yellow] Recipe piperider command failed: '{cmd}' with exit code: {exit_code}")
            sys.exit(exit_code)
        console.print()


def get_current_branch(cfg: RecipeConfiguration):
    """
    Update the effective branch name for cfg and return the original branch before execution
    """

    if cfg.base.branch is None and cfg.target.branch is None:
        # We don't care the current branch, because we won't change it
        return None

    original_branch = tool().git_branch()

    return original_branch


def switch_merge_base_branch(a: str, b: str) -> str:
    base_commit = tool().git_merge_base(a, b)
    console.print(f"Switch git branch to: \[merge-base of {a}...{b} -> {base_commit}]")
    tool().git_checkout_to(base_commit)
    return base_commit


def switch_branch(branch_name):
    tool().git_checkout_to(branch_name)


def clean_up(cfg: RecipeConfiguration):
    if cfg.base.tmp_dir_path:
        console.print(f"Clean up base branch git archive: \[{cfg.base.tmp_dir_path}]")
        tool().remove_dir(cfg.base.tmp_dir_path)

    if cfg.target.tmp_dir_path:
        console.print(f"Clean up base branch git archive: \[{cfg.target.tmp_dir_path}]")
        tool().remove_dir(cfg.target.tmp_dir_path)


def execute_recipe_configuration(cfg: RecipeConfiguration, select: tuple = None, modified: bool = False, debug=False):
    console.rule("Recipe executor: verify execution environments")
    # check the dependencies
    console.print("Check: git")
    verify_git_dependencies(cfg)
    console.print("Check: dbt")
    verify_dbt_dependencies(cfg)

    try:
        console.rule("Recipe executor: prepare execution environments")
        dbt_resources, dbt_state_path = prepare_dbt_resources_candidate(cfg, select=select, modified=modified)
        if dbt_resources:
            if debug:
                console.print(f'Config: piperider env "PIPERIDER_DBT_RESOURCES" = {dbt_resources}')
            else:
                console.print('Config: piperider env "PIPERIDER_DBT_RESOURCES"')
            cfg.base.piperider.environments['PIPERIDER_DBT_RESOURCES'] = '\n'.join(dbt_resources)
            cfg.target.piperider.environments['PIPERIDER_DBT_RESOURCES'] = '\n'.join(dbt_resources)

        console.rule("Recipe executor: base phase")
        execute_recipe_archive(cfg.base, recipe_type='base', debug=debug)

        console.rule("Recipe executor: target phase")
        execute_recipe(cfg.target, recipe_type='target', debug=debug)

    except Exception as e:
        if isinstance(e, InteractiveStopException):
            console.rule("Recipe executor: interrupted by the user", style="red")
            sys.exit(0)
        else:
            console.rule("Recipe executor: error occurred", style="red")
            raise e


def select_recipe_file(name: str = None):
    if name is None:
        name = "default"
    recipe_path = os.path.join(FileSystem.PIPERIDER_RECIPES_PATH, f"{name}.yml")
    if not os.path.exists(recipe_path):
        if name == "default":
            return None
        else:
            raise FileNotFoundError(
                f"Recipe file not found: {recipe_path}",
            )

    return recipe_path


@dataclass
class ExecutionFlags:
    dry_run: bool
    interactive: bool


_execution_flags = ExecutionFlags(dry_run=False, interactive=False)


def is_recipe_dry_run():
    global _execution_flags
    return _execution_flags.dry_run


def is_recipe_interactive():
    global _execution_flags
    return _execution_flags.interactive


def configure_recipe_execution_flags(dry_run: bool, interactive: bool):
    global _execution_flags, console
    _execution_flags = ExecutionFlags(dry_run=dry_run is True, interactive=interactive is True)
    if _execution_flags.dry_run:
        console = Console()
        console.original_print = console.print

        def dry_run_print(*objects: Any):
            if objects:
                if isinstance(objects[0], str):
                    console.original_print("\[dry-run]", end=' ')
                console.original_print(*objects)
            else:
                console.original_print()

        console.print = dry_run_print
    else:
        console = Console()

    if is_recipe_dry_run() and is_recipe_interactive():
        console.print("[yellow]Warnings[/yellow]: in dry-run mode, we will ignore --interactive.")


configure_recipe_execution_flags(dry_run=False, interactive=False)


def tool():
    from piperider_cli.recipes.utils import (DryRunRecipeUtils,
                                             InteractiveRecipeDecorator,
                                             RecipeUtils)
    if is_recipe_dry_run():
        return DryRunRecipeUtils(console)
    else:
        if is_recipe_interactive():
            return InteractiveRecipeDecorator(RecipeUtils(console))
        else:
            return RecipeUtils(console)


if __name__ == '__main__':

    test_recipe_path = os.path.join(os.path.dirname(__file__), 'example_recipe.yml')
    try:

        recipe = RecipeConfiguration.load(test_recipe_path)
        console.print(recipe.__dict__())
    except ValidationError as e:
        print(e.message)
    exit(1)
